# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import prettytable

from .. import coresight
from ..core import exceptions
from ..probe.debug_probe import DebugProbe
from ..coresight.ap import MEM_AP
from ..core.target import Target
from ..utility.cmdline import (
    convert_session_options,
    convert_frequency,
    convert_vector_catch,
    )
from .base import ValueBase

LOG = logging.getLogger(__name__)

VC_NAMES_MAP = {
        Target.VectorCatch.HARD_FAULT: "hard fault",
        Target.VectorCatch.BUS_FAULT: "bus fault",
        Target.VectorCatch.MEM_FAULT: "memory fault",
        Target.VectorCatch.INTERRUPT_ERR: "interrupt error",
        Target.VectorCatch.STATE_ERR: "state error",
        Target.VectorCatch.CHECK_ERR: "check error",
        Target.VectorCatch.COPROCESSOR_ERR: "coprocessor error",
        Target.VectorCatch.CORE_RESET: "core reset",
        Target.VectorCatch.SECURE_FAULT: "secure fault",
        }

HPROT_BIT_DESC = {
        0: ("instruction fetch", "data access"),
        1: ("user", "privileged"),
        2: ("non-bufferable", "bufferable"),
        3: ("non-cacheable", "cacheable/modifiable"),
        4: ("no cache lookup", "lookup in cache"),
        5: ("no cache allocate", "allocate in cache"),
        6: ("non-shareable", "shareable"),
        }

class UniqueIdValue(ValueBase):
    INFO = {
            'names': ['probe-uid', 'uid'],
            'group': 'standard',
            'category': 'probe',
            'access': 'r',
            'help': "Target's unique ID.",
            }

    def display(self, args):
        self.context.writei("Unique ID:    %s", self.context.board.unique_id)

class TargetValue(ValueBase):
    INFO = {
            'names': ['target'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "General target information.",
            }

    def display(self, args):
        self.context.writei("Target:       %s", self.context.target.part_number)
        self.context.writei("DAP IDCODE:   0x%08x", self.context.target.dp.dpidr.idr)

class CoresValue(ValueBase):
    INFO = {
            'names': ['cores'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "Information about CPU cores in the target.",
            }

    def display(self, args):
        if self.context.target.is_locked():
            self.context.write("Target is locked")
        else:
            self.context.writei("Cores:        %d", len(self.context.target.cores))
            for i, core in self.context.target.cores.items():
                self.context.writei("Core %d type:  %s%s", i,
                        coresight.core_ids.CORE_TYPE_NAME[core.core_type],
                        " (selected)" if ((self.context.selected_core is not None) \
                                            and (self.context.selected_core.core_number == i)) else "")

class APsValue(ValueBase):
    INFO = {
            'names': ['aps'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "List discovered Access Ports.",
            }

    def display(self, args):
        if self.context.target.is_locked():
            self.context.write("Target is locked")
        else:
            self.context.writei("%d APs:", len(self.context.target.aps))
            for addr, ap in sorted(self.context.target.aps.items(), key=lambda x: x[0]):
                self.context.writei("%s: %s%s", addr, ap.type_name,
                        " (selected)" if (self.context.selected_ap_address == addr) else "")

class MemoryMapValue(ValueBase):
    INFO = {
            'names': ['map'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "Target memory map.",
            }

    def display(self, args):
        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return

        pt = prettytable.PrettyTable(["Region", "Type", "Start", "End", "Size", "Access", "Sector", "Page"])
        pt.align = 'l'
        pt.border = False
        for region in self.context.selected_core.get_memory_map():
            pt.add_row([
                region.name,
                region.type.name.capitalize(),
                "0x%08x" % region.start,
                "0x%08x" % region.end,
                "0x%08x" % region.length,
                region.access,
                ("0x%08x" % region.sector_size) if region.is_flash else '-',
                ("0x%08x" % region.page_size) if region.is_flash else '-',
                ])
        self.context.write(pt)

class PeripheralsValue(ValueBase):
    INFO = {
            'names': ['peripherals'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "List of target peripheral instances.",
            }

    def display(self, args):
        for periph in sorted(self.context.peripherals.values(), key=lambda x:x.base_address):
            self.context.writei("0x%08x: %s", periph.base_address, periph.name)

class FaultValue(ValueBase):
    INFO = {
            'names': ['fault'],
            'group': 'standard',
            'category': 'exceptions',
            'access': 'r',
            'show_usage': "[-a]",
            'help': "Fault status information.",
            'extra_help': "By default, only asserted fields are shown. Add -a to command to show all fields.",
            }

    def display(self, args):
        showAll = ('-a' in args)
        
        CFSR = 0xe000ed28
        HFSR = 0xe000ed2c
        DFSR = 0xe000ed30
        MMFAR = 0xe000ed34
        BFAR = 0xe000ed38
#         AFSR = 0xe000ed3c
        
        MMFSR_fields = [
                ('IACCVIOL', 0),
                ('DACCVIOL', 1),
                ('MUNSTKERR', 3),
                ('MSTKERR', 4),
#                 ('MMARVALID', 7),
            ]
        BFSR_fields = [
                ('IBUSERR', 0),
                ('PRECISERR', 1),
                ('IMPRECISERR', 2),
                ('UNSTKERR', 3),
                ('STKERR', 4),
                ('LSPERR', 5),
#                 ('BFARVALID', 7),
            ]
        UFSR_fields = [
                ('UNDEFINSTR', 0),
                ('INVSTATE', 1),
                ('INVPC', 2),
                ('NOCP', 3),
                ('STKOF', 4),
                ('UNALIGNED', 8),
                ('DIVBYZERO', 9),
            ]
        HFSR_fields = [
                ('VECTTBL', 1),
                ('FORCED', 30),
                ('DEBUGEVT', 31),
            ]
        DFSR_fields = [
                ('HALTED', 0),
                ('BKPT', 1),
                ('DWTTRAP', 2),
                ('VCATCH', 3),
                ('EXTERNAL', 4),
            ]
        
        def print_fields(regname, value, fields, showAll):
            if value == 0 and not showAll:
                return
            self.context.writei("  %s = 0x%08x", regname, value)
            for name, bitpos in fields:
                bit = (value >> bitpos) & 1
                if showAll or bit != 0:
                    self.context.writei("    %s = 0x%x", name, bit)
        
        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return

        cfsr = self.context.selected_core.read32(CFSR)
        mmfsr = cfsr & 0xff
        bfsr = (cfsr >> 8) & 0xff
        ufsr = (cfsr >> 16) & 0xffff
        hfsr = self.context.selected_core.read32(HFSR)
        dfsr = self.context.selected_core.read32(DFSR)
        mmfar = self.context.selected_core.read32(MMFAR)
        bfar = self.context.selected_core.read32(BFAR)
        
        print_fields('MMFSR', mmfsr, MMFSR_fields, showAll)
        if showAll or mmfsr & (1 << 7): # MMFARVALID
            self.context.writei("  MMFAR = 0x%08x", mmfar)
        print_fields('BFSR', bfsr, BFSR_fields, showAll)
        if showAll or bfsr & (1 << 7): # BFARVALID
            self.context.writei("  BFAR = 0x%08x", bfar)
        print_fields('UFSR', ufsr, UFSR_fields, showAll)
        print_fields('HFSR', hfsr, HFSR_fields, showAll)
        print_fields('DFSR', dfsr, DFSR_fields, showAll)

class NresetValue(ValueBase):
    INFO = {
            'names': ['nreset'],
            'group': 'standard',
            'category': 'target',
            'access': 'rw',
            'help': "Current nRESET signal state.",
            'extra_help': "Accepts a value of 0 or 1.",
            }

    def display(self, args):
        rst = int(not self.context.probe.is_reset_asserted())
        self.context.writef("nRESET = {}", rst)

    def modify(self, args):
        if len(args) != 1:
            raise exceptions.CommandError("missing reset state")
        state = int(args[0], base=0)
        self.context.writef("nRESET = {}", state)
        
        # Use the probe to assert reset if the DP doesn't exist for some reason, otherwise
        # use the DP so reset notifications are sent.
        if self.context.target.dp is None:
            self.context.probe.assert_reset((state == 0))
        else:
            self.context.target.dp.assert_reset((state == 0))

class SessionOptionValue(ValueBase):
    INFO = {
            'names': ['option'],
            'group': 'standard',
            'category': 'options',
            'access': 'rw',
            'show_usage': "NAME+",
            'set_usage': "NAME[=VALUE]+",
            'help': "The current value of one or more session options.",
            'extra_help': "When setting, each argument should follow the form \"NAME[=VALUE]\".",
            }

    def display(self, args):
        if len(args) < 1:
            raise exceptions.CommandError("missing session option name argument")
        for name in args:
            try:
                value = self.context.session.options[name]
                self.context.writei("Option '%s' = %s", name, value)
            except KeyError:
                self.context.writei("No option with name '%s'", name)

    def modify(self, args):
        if len(args) < 1:
            raise exceptions.CommandError("missing session option setting")
        opts = convert_session_options(args)
        self.context.session.options.update(opts)

class MemApValue(ValueBase):
    INFO = {
            'names': ['mem-ap'],
            'group': 'standard',
            'category': 'memory',
            'access': 'rw',
            'help': "The currently selected MEM-AP used for memory read/write commands.",
            'extra_help': "When the selected core is changed by the 'core' command, the selected "
                "MEM-AP is changed to match. This overrides a user-selected MEM-AP if different "
                "from the AP for the newly selected core.",
            }

    def display(self, args):
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        self.context.writef("{} is selected", self.context.selected_ap.short_description)

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing argument")
            
        ap_num = int(args[0], base=0)
        if self.context.target.dp.adi_version == coresight.dap.ADIVersion.ADIv5:
            ap_addr = coresight.ap.APv1Address(ap_num)
        elif self.context.target.dp.adi_version == coresight.dap.ADIVersion.ADIv6:
            ap_addr = coresight.ap.APv2Address(ap_num)
        if ap_addr not in self.context.target.aps:
            self.context.writef("Invalid AP number {:#x}", ap_num)
            return
        ap = self.context.target.aps[ap_addr]
        if not isinstance(ap, MEM_AP):
            self.context.writef("{} is not a MEM-AP", ap.short_description)
            return
        self.context.selected_ap_address = ap_addr

class HnonsecValue(ValueBase):
    INFO = {
            'names': ['hnonsec'],
            'group': 'standard',
            'category': 'memory',
            'access': 'rw',
            'help': "The current HNONSEC value used by the selected MEM-AP.",
            }

    def display(self, args):
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        self.context.writef("{} HNONSEC = {} ({})",
            self.context.selected_ap.short_description,
            self.context.selected_ap.hnonsec,
            ("nonsecure" if self.context.selected_ap.hnonsec else "secure"))

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing argument")
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        value = int(args[0], base=0)
        self.context.selected_ap.hnonsec = value

class HprotValue(ValueBase):
    INFO = {
            'names': ['hprot'],
            'group': 'standard',
            'category': 'memory',
            'access': 'rw',
            'help': "The current HPROT value used by the selected MEM-AP.",
            }

    def display(self, args):
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        hprot = self.context.selected_ap.hprot
        self.context.writef("{} HPROT = {:#x}",
            self.context.selected_ap.short_description,
            hprot)
        desc = ""
        for bitnum in range(7):
            bitvalue = (hprot >> bitnum) & 1
            desc += "    HPROT[{}] = {:#x} ({})\n".format(
                bitnum,
                bitvalue,
                HPROT_BIT_DESC[bitnum][bitvalue])
        self.context.write(desc, end='')

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing argument")
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        value = int(args[0], base=0)
        self.context.selected_ap.hprot = value

class TargetGraphValue(ValueBase):
    INFO = {
            'names': ['graph'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "Print the target object graph.",
            }

    def display(self, args):
        self.context.board.dump()
    
class LockedValue(ValueBase):
    INFO = {
            'names': ['locked'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "Report whether the target is locked.",
            }

    def display(self, args):
        if self.context.target.is_locked():
            self.context.write("Taget is locked")
        else:
            self.context.write("Taget is unlocked")

class RegisterGroupsValue(ValueBase):
    INFO = {
            'names': ['register-groups'],
            'group': 'standard',
            'category': 'registers',
            'access': 'r',
            'help': "Display available register groups for the selected core.",
            }

    def display(self, args):
        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return
        for g in sorted(self.context.selected_core.core_registers.groups):
            self.context.write(g)

class VectorCatchValue(ValueBase):
    INFO = {
            'names': ['vector-catch', 'vc'],
            'group': 'standard',
            'category': 'exceptions',
            'access': 'rw',
            'help': "Show current vector catch settings.",
            'extra_help': "When setting, the alue is a concatenation of one letter per enabled source in "
                "any order, or 'all' or 'none'. (h=hard fault, b=bus fault, m=mem fault, i=irq err, s=state "
                "err, c=check err, p=nocp, r=reset, a=all, n=none).",
            }

    def display(self, args):
        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return

        catch = self.context.selected_core.get_vector_catch()

        self.context.write("Vector catch:")
        for mask in sorted(VC_NAMES_MAP.keys()):
            name = VC_NAMES_MAP[mask]
            s = "ON" if (catch & mask) else "OFF"
            self.context.writef("  {:3} {}", s, name)

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing vector catch setting")

        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return
    
        try:
            self.context.selected_core.set_vector_catch(convert_vector_catch(args[0]))
        except ValueError as e:
            self.context.write(e)

class StepIntoInterruptsValue(ValueBase):
    INFO = {
            'names': ['step-into-interrupts', 'si'],
            'group': 'standard',
            'category': 'options',
            'access': 'rw',
            'help': "Display whether interrupts are enabled when single stepping.",
            'extra_help': "Set to 1 to enable.",
            }

    def display(self, args):
        self.context.writei("Interrupts while stepping: %s",
            ("enabled" if self.context.session.options['step_into_interrupt'] else "disabled"))

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing argument")
        value = (args[0] in ('1', 'true', 'yes', 'on'))
        self.context.session.options['step_into_interrupt'] = value

class LogLevelValue(ValueBase):
    INFO = {
            'names': ['log'],
            'group': 'standard',
            'category': 'miscellaneous',
            'access': 'w',
            'set_usage': "LEVEL [MODULE+]",
            'help': "Set log level to one of 'debug', 'info', 'warning', 'error', 'critical'.",
            'extra_help': "If pyocd module names are provided as arguments after the log level then "
                            "only those modules will have their log level changed."
            }

    LEVELS = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL,
            }

    def display(self, args):
        pass

    def modify(self, args):
        if len(args) < 1:
            raise exceptions.CommandError("no log level provided")
        level_name = args[0].lower()
        if level_name not in self.LEVELS:
            raise exceptions.CommandError("log level must be one of {%s}" % ','.join(self.LEVELS.keys()))
        level = self.LEVELS[level_name]
        if len(args) == 1:
            logging.getLogger().setLevel(level)
        else:
            for module_name in args[1:]:
                # Check if there is already a logger with this name.
                #
                # This is guarded with a check that the root logger has a 'manager' attribute, and the
                # the Manager object has a 'loggerDict' attribute. Both attributes are undocumented, so
                # a check is necessary to prevent failure in case the logging code changes.
                if hasattr(logging.root, 'manager'):
                    manager = logging.root.manager
                    if hasattr(manager, 'loggerDict') and (module_name not in manager.loggerDict):
                        self.context.writef("No logger exists with the name {}", module_name)
                        continue
                logging.getLogger(module_name).setLevel(level)

class ClockFrequencyValue(ValueBase):
    INFO = {
            'names': ['frequency'],
            'group': 'standard',
            'category': 'options',
            'access': 'w',
            'help': "Set SWD or JTAG clock frequency in Hertz.",
            'extra_help': "A case-insensitive metric scale suffix of either 'k' or 'm' is allowed, as well "
                "as a trailing \"Hz\". There must be no space between the frequency and the suffix. For "
                "example, \"2.5MHz\" sets the clock to 2.5 MHz.",
            }

    def display(self, args):
        pass

    def modify(self, args):
        if len(args) < 1:
            raise exceptions.CommandError("no clock frequency provided")
        try:
            freq_Hz = convert_frequency(args[0])
        except:
            raise exceptions.CommandError("invalid frequency")
        self.context.probe.set_clock(freq_Hz)
        if self.context.probe.wire_protocol == DebugProbe.Protocol.SWD:
            swd_jtag = 'SWD'
        elif self.context.probe.wire_protocol == DebugProbe.Protocol.JTAG:
            swd_jtag = 'JTAG'
        else:
            swd_jtag = '??'

        if freq_Hz >= 1000000:
            nice_freq = "%.2f MHz" % (freq_Hz / 1000000)
        elif freq_Hz > 1000:
            nice_freq = "%.2f kHz" % (freq_Hz / 1000)
        else:
            nice_freq = "%d Hz" % freq_Hz

        self.context.writei("Changed %s frequency to %s", swd_jtag, nice_freq)




# pyOCD debugger
# Copyright (c) 2015-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core.target import Target
from .component import CoreSightComponent
from ..debug.breakpoints.provider import (Breakpoint, BreakpointProvider)
import logging

LOG = logging.getLogger(__name__)

class HardwareBreakpoint(Breakpoint):
    def __init__(self, comp_register_addr, provider):
        super(HardwareBreakpoint, self).__init__(provider)
        self.comp_register_addr = comp_register_addr
        self.type = Target.BreakpointType.HW

class FPB(BreakpointProvider, CoreSightComponent):
    """! @brief Flash Patch and Breakpoint unit"""
    
    # FPB registers
    #
    # The addresses are offsets from the base address.
    FP_CTRL = 0x00000000
    FP_CTRL_KEY = 1 << 1
    FP_CTRL_REV_MASK = 0xf0000000
    FP_CTRL_REV_SHIFT = 28
    FP_COMP0 = 0x00000008
    
    def __init__(self, ap, cmpid=None, addr=None):
        CoreSightComponent.__init__(self, ap, cmpid, addr)
        BreakpointProvider.__init__(self)
        self.hw_breakpoints = []
        self.nb_code = 0
        self.nb_lit = 0
        self.num_hw_breakpoint_used = 0
        self.enabled = False
        self.fpb_rev = 1

    @property
    def revision(self):
        return self.fpb_rev

    def init(self):
        """! @brief Inits the FPB.
        
        Reads the number of hardware breakpoints available on the core and disable the FPB
        (Flash Patch and Breakpoint Unit), which will be enabled when the first breakpoint is set.
        setup FPB (breakpoint)
        """
        fpcr = self.ap.read32(self.address + FPB.FP_CTRL)
        self.fpb_rev = 1 + ((fpcr & FPB.FP_CTRL_REV_MASK) >> FPB.FP_CTRL_REV_SHIFT)
        if self.fpb_rev not in (1, 2):
            LOG.warning("Unknown FPB version %d", self.fpb_rev)
        self.nb_code = ((fpcr >> 8) & 0x70) | ((fpcr >> 4) & 0xF)
        self.nb_lit = (fpcr >> 7) & 0xf
        LOG.info("%d hardware breakpoints, %d literal comparators", self.nb_code, self.nb_lit)
        for i in range(self.nb_code):
            self.hw_breakpoints.append(HardwareBreakpoint(self.address + FPB.FP_COMP0 + 4*i, self))

        # disable FPB (will be enabled on first bp set)
        self.disable()
        for bp in self.hw_breakpoints:
            self.ap.write_memory(bp.comp_register_addr, 0)

    @property
    def bp_type(self):
        return Target.BreakpointType.HW

    def enable(self):
        self.ap.write_memory(self.address + FPB.FP_CTRL, FPB.FP_CTRL_KEY | 1)
        self.enabled = True
        LOG.debug('fpb has been enabled')
        return

    def disable(self):
        self.ap.write_memory(self.address + FPB.FP_CTRL, FPB.FP_CTRL_KEY | 0)
        self.enabled = False
        LOG.debug('fpb has been disabled')
        return

    @property
    def available_breakpoints(self):
        return len(self.hw_breakpoints) - self.num_hw_breakpoint_used

    def can_support_address(self, addr):
        """! @brief Test whether an address is supported by the FPB.
        
        For FPBv1, hardware breakpoints are only supported in the range 0x00000000 - 0x1fffffff.
        This was fixed for FPBv2, which supports hardware breakpoints at any address.
        """
        return (self.fpb_rev == 2) or (addr < 0x20000000)

    def set_breakpoint(self, addr):
        """! @brief Set a hardware breakpoint at a specific location in flash."""
        if not self.enabled:
            self.enable()

        if not self.can_support_address(addr):
            LOG.error('Breakpoint out of range 0x%X', addr)
            return None

        if self.available_breakpoints == 0:
            LOG.error('No more hardware breakpoints are available, dropped breakpoint at 0x%08x', addr)
            return None

        for bp in self.hw_breakpoints:
            if not bp.enabled:
                bp.enabled = True
                comp = 0
                if self.fpb_rev == 1:
                    bp_match = (1 << 30)
                    if addr & 0x2:
                        bp_match = (2 << 30)
                    comp = addr & 0x1ffffffc | bp_match | 1
                elif self.fpb_rev == 2:
                    comp = (addr & 0xfffffffe) | 1
                self.ap.write32(bp.comp_register_addr, comp)
                LOG.debug("BP: wrote 0x%08x to comp @ 0x%08x", comp, bp.comp_register_addr)
                bp.addr = addr
                self.num_hw_breakpoint_used += 1
                return bp
        return None

    def remove_breakpoint(self, bp):
        """! @brief Remove a hardware breakpoint at a specific location in flash."""
        for hwbp in self.hw_breakpoints:
            if hwbp.enabled and hwbp.addr == bp.addr:
                hwbp.enabled = False
                self.ap.write_memory(hwbp.comp_register_addr, 0)
                self.num_hw_breakpoint_used -= 1
                return




# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import argparse
import sys
import traceback
import logging
import tempfile

from pyocd.core.helpers import ConnectHelper
from pyocd.probe.pydapaccess import DAPAccess
from pyocd.utility.mask import round_up_div
from pyocd.utility import conversion
from pyocd.core.memory_map import MemoryType
from pyocd.commands.execution_context import CommandExecutionContext
from pyocd.commands import commands
from test_util import (
    Test,
    TestResult,
    get_session_options,
    get_target_test_params,
    binary_to_hex_file,
    binary_to_elf_file,
    get_test_binary_path,
    )

class CommandsTestResult(TestResult):
    def __init__(self):
        super(CommandsTestResult, self).__init__(None, None, None)
        self.name = "commands"

class CommandsTest(Test):
    def __init__(self):
        super(CommandsTest, self).__init__("Commands Test", commands_test)

    def run(self, board):
        try:
            result = self.test_function(board.unique_id)
        except Exception as e:
            result = CommandsTestResult()
            result.passed = False
            print("Exception %s when testing board %s" % (e, board.unique_id))
            traceback.print_exc(file=sys.stdout)
        result.board = board
        result.test = self
        return result

def commands_test(board_id):
    with ConnectHelper.session_with_chosen_probe(unique_id=board_id, **get_session_options()) as session:
        board = session.board
        target = session.target
        target_type = board.target_type

        test_params = get_target_test_params(session)
        session.probe.set_clock(test_params['test_clock'])

        memory_map = board.target.get_memory_map()
        boot_region = memory_map.get_boot_memory()
        ram_region = memory_map.get_default_region_of_type(MemoryType.RAM)
        ram_base = ram_region.start
        boot_start_addr = boot_region.start
        boot_end_addr = boot_region.end
        boot_blocksize = boot_region.blocksize
        binary_file = get_test_binary_path(board.test_binary)

        # Generate an Intel hex file from the binary test file.
        temp_test_hex_name = binary_to_hex_file(binary_file, boot_region.start)

        temp_bin_file = tempfile.mktemp('.bin')
        
        with open(binary_file, "rb") as f:
            test_data = list(bytearray(f.read()))
        test_data_length = len(test_data)
        test_file_sectors = round_up_div(test_data_length, boot_blocksize)
        boot_first_free_block_addr = boot_start_addr + test_file_sectors * boot_blocksize
        reset_handler_addr = conversion.byte_list_to_u32le_list(test_data[4:8])[0]

        test_pass_count = 0
        test_count = 0
        failed_commands = []
        result = CommandsTestResult()
        
        context = CommandExecutionContext()
        context.attach_session(session)
        
        COMMANDS_TO_TEST = [
                "status",
                "reset",
                "reset halt",
                "reg",
                "reg general",
                "reg all",
                "reg r0",
                "wreg r0 0x12345678",
#                 "d pc", # Disable disasm because capstone is not installed by default.
#                 "d --center pc 32",
                "read32 0x%08x" % (boot_start_addr + boot_blocksize),
                "read16 0x%08x" % (boot_start_addr + boot_blocksize),
                "read8 0x%08x" % (boot_start_addr + boot_blocksize),
                "rw 0x%08x 16" % ram_base,
                "rh 0x%08x 16" % ram_base,
                "rb 0x%08x 16" % ram_base,
                "write32 0x%08x 0x11223344 0x55667788" % ram_base,
                "write16 0x%08x 0xabcd" % (ram_base + 8),
                "write8 0x%08x 0 1 2 3 4 5 6" % (ram_base + 10),
                "savemem 0x%08x 128 '%s'" % (boot_start_addr, temp_bin_file),
                "loadmem 0x%08x '%s'" % (ram_base, temp_bin_file),
                "loadmem 0x%08x '%s'" % (boot_start_addr, binary_file),
                "load '%s'" % temp_test_hex_name,
                "load '%s' 0x%08x" % (binary_file, boot_start_addr),
                "compare 0x%08x '%s'" % (ram_base, temp_bin_file),
                "compare 0x%08x 32 '%s'" % (ram_base, temp_bin_file),
                "fill 0x%08x 128 0xa5" % ram_base,
                "fill 16 0x%08x 64 0x55aa" % (ram_base + 64),
                "find 0x%08x 128 0xaa 0x55" % ram_base, # find that will pass
                "find 0x%08x 128 0xff" % ram_base, # find that will fail
                "erase 0x%08x" % (boot_first_free_block_addr),
                "erase 0x%08x 1" % (boot_first_free_block_addr + boot_blocksize),
                "go",
                "halt",
                "step",
                "s 4",
                "continue",
                "h",
                "break 0x%08x" % reset_handler_addr,
                "lsbreak",
                "rmbreak 0x%08x" % reset_handler_addr,
                "watch 0x%08x" % ram_base,
                "lswatch",
                "rmwatch 0x%08x" % ram_base,
                "watch 0x%08x rw 2" % ram_base,
                "rmwatch 0x%08x" % ram_base,
                "core",
                "core 0",
                "readdp 0", # read DPIDR
                "writedp 0 0x1e", # write ABORT to clear error flags
                "readap 0xfc", # read IDR
                "readap 0 0xfc", # read IDR of AP#0
                "writeap 0x4 0", # set TAR to 0
                "writeap 0 0x4 0", # set TAR to 0 on AP#0
                "gdbserver start",
                "gdbserver status",
                "gdbserver stop",
                "probeserver start",
                "probeserver status",
                "probeserver stop",
                "show probe-uid",
                "show target",
                "show cores",
                "show map",
                "show peripherals",
                "show fault",
                "show nreset",
                "set nreset 1",
                "show option reset_type",
                "set option reset_type=sw",
                "show mem-ap",
                "set mem-ap 0",
                "show hnonsec",
                "set hnonsec 0",
                "show hprot",
                "set hprot 0x3", # set default hprot: data, priv
                "show graph",
                "show locked",
                "show register-groups",
                "show vector-catch",
                "set vector-catch all",
                "show step-into-interrupts",
                "set step-into-interrupts 1",
                "set log info",
                "set frequency %d" % test_params['test_clock'],
                
                # Semicolon-separated commands.
                'rw 0x%08x ; rw 0x%08x' % (ram_base, ram_base + 4),
                
                # Python and system commands.
                '$2+ 2',
                '!echo hello',
                '!echo hi \; echo there', # using escaped semicolon in a sytem command

                # Commands not tested:
#                 "list",
#                 "erase", # chip erase
#                 "unlock",
#                 "exit",
#                 "initdp",
#                 "makeap",
#                 "reinit",
#                 "where",
#                 "symbol",
                ]
        
        # For now we just verify that the commands run without raising an exception.
        print("\n------ Testing commands ------")
        
        def test_command(cmd):
            try:
                print("\nTEST: %s" % cmd)
                context.process_command_line(cmd)
            except:
                print("TEST FAILED")
                failed_commands.append(cmd)
                traceback.print_exc(file=sys.stdout)
                return False
            else:
                print("TEST PASSED")
                return True

        for cmd in COMMANDS_TO_TEST:
            if test_command(cmd):
                test_pass_count += 1
            test_count += 1

        print("\n\nTest Summary:")
        print("Pass count %i of %i tests" % (test_pass_count, test_count))
        if failed_commands:
            for c in failed_commands:
                print(" - '" + c + "'")
        if test_pass_count == test_count:
            print("COMMANDS TEST SCRIPT PASSED")
        else:
            print("COMMANDS TEST SCRIPT FAILED")

        target.reset()

        result.passed = test_count == test_pass_count
        return result

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='pyOCD commands test')
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    parser.add_argument('-u', '--uid', help='Debug probe unique ID')
    parser.add_argument("-da", "--daparg", dest="daparg", nargs='+', help="Send setting to DAPAccess layer.")
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    DAPAccess.set_args(args.daparg)
    commands_test(args.uid)




# pyOCD debugger
# Copyright (c) 2006-2013 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (RamRegion, MemoryMap)

class RTL8195AM(CoreSightTarget):

    VENDOR = "Realtek Semiconductor"
    
    MEMORY_MAP = MemoryMap(
        RamRegion(      start=0x00000000,  length=0x400000),
        RamRegion(      start=0x10000000,  length=0x80000),
        RamRegion(      start=0x30000000,  length=0x200000),
        RamRegion(      start=0x40000000,  length=0x40000)
        )

    def __init__(self, session):
        super(RTL8195AM, self).__init__(session, self.MEMORY_MAP)



# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (c) 2020 Patrick Huesmann
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
from time import sleep

from .sink import TraceEventSink
from .events import TraceITMEvent
from .swo import SWOParser
from ..coresight.itm import ITM
from ..coresight.tpiu import TPIU
from ..core.target import Target
from ..core import exceptions
from ..probe.debug_probe import DebugProbe
from ..utility.server import StreamServer

LOG = logging.getLogger(__name__)

class SWVEventSink(TraceEventSink):
    """! @brief Trace event sink that converts ITM packets to a text stream."""
    
    def __init__(self, console):
        """! @brief Constructor.
        @param self
        @param console File-like object to which SWV data will be written.
        """
        self._console = console
    
    def receive(self, event):
        """! @brief Handle an SWV trace event.
        @param self
        @param event An instance of TraceITMEvent. If the event is not this class, or isn't
            for ITM port 0, then it will be ignored. The individual bytes of 16- or 32-bit events
            will be extracted and written to the console.
        """
        if not isinstance(event, TraceITMEvent):
            return

        # Extract bytes.
        if event.width == 1:
            data = chr(event.data)
        elif event.width == 2:
            data = chr(event.data & 0xff) + chr((event.data >> 8) & 0xff)
        elif event.width == 4:
            data = (chr(event.data & 0xff)
                    + chr((event.data >> 8) & 0xff)
                    + chr((event.data >> 16) & 0xff)
                    + chr((event.data >> 24) & 0xff))

        self._console.write(data)

class SWVReader(threading.Thread):
    """! @brief Sets up SWV and processes data in a background thread."""

    def __init__(self, session, core_number=0, lock=None):
        """! @brief Constructor.
        @param self
        @param session The Session instance.
        @param core_number The number of the core being traced. Default is core 0.
        """
        super(SWVReader, self).__init__()
        self.name = "SWVReader"
        self.daemon = True
        self._session = session
        self._core_number = core_number
        self._shutdown_event = threading.Event()
        self._swo_clock = 0
        self._lock = lock
        
        self._session.subscribe(self._reset_handler, Target.Event.POST_RESET, self._session.target.cores[core_number])
        
    def init(self, sys_clock, swo_clock, console):
        """! @brief Configures trace graph and starts thread.
        
        This method performs all steps required to start up SWV. It first calls the target's
        trace_start() method, which allows for target-specific trace initialization. Then it
        configures the TPIU and ITM modules. A simple trace data processing graph is created that
        connects an SWVEventSink with a SWOParser. Finally, the reader thread is started.
        
        If the debug probe does not support SWO, a warning is printed but nothing else is done.
        
        @param self
        @param sys_clock
        @param swo_clock
        @param console
        """
        self._swo_clock = swo_clock
        
        if DebugProbe.Capability.SWO not in self._session.probe.capabilities:
            LOG.warning("Probe %s does not support SWO", self._session.probe.unique_id)
            return
        
        self._session.target.trace_start()
        
        itm = self._session.target.get_first_child_of_type(ITM)
        tpiu = self._session.target.get_first_child_of_type(TPIU)

        itm.init()
        itm.enable()
        tpiu.init()

        if tpiu.set_swo_clock(swo_clock, sys_clock):
            LOG.info("Set SWO clock to %d", swo_clock)
        else:
            LOG.warning("Failed to set SWO clock rate")
            return

        self._parser = SWOParser(self._session.target.cores[self._core_number])
        self._sink = SWVEventSink(console)
        self._parser.connect(self._sink)
        
        self.start()
    
    def stop(self):
        """! @brief Stops processing SWV data.
        
        The reader thread is terminated first, then the ITM is disabled. The last step is to call
        the target's trace_stop() method.
        
        Does nothing if the init() method did not complete successfully.
        """
        if not self.is_alive():
            return

        self._shutdown_event.set()
        self.join()

        itm = self._session.target.get_first_child_of_type(ITM)
        itm.disable()
        
        self._session.target.trace_stop()
    
    def run(self):
        """! @brief SWV reader thread routine.
        
        Starts the probe receiving SWO data by calling DebugProbe.swo_start(). For as long as the
        thread runs, it reads SWO data from the probe and passes it to the SWO parser created in
        init(). When the thread is signaled to stop, it calls DebugProbe.swo_stop() before exiting.
        """

        if self._lock:
            self._lock.acquire()

        swv_raw_server = StreamServer(
                            self._session.options.get('swv_raw_port'),
                            serve_local_only=self._session.options.get('serve_local_only'),
                            name="SWV raw",
                            is_read_only=True) \
                         if self._session.options.get('swv_raw_enable') else None

        # Stop SWO first in case the probe already had it started. Ignore if this fails.
        try:
            self._session.probe.swo_stop()
        except exceptions.ProbeError:
            pass
        self._session.probe.swo_start(self._swo_clock)
        
        while not self._shutdown_event.is_set():
            data = self._session.probe.swo_read()
            if data:
                if swv_raw_server:
                    swv_raw_server.write(data)
                self._parser.parse(data)
        
            if self._lock:
                self._lock.release()

            sleep(0.001)

            if self._lock:
                self._lock.acquire()
            
        self._session.probe.swo_stop()

        if swv_raw_server:
            swv_raw_server.stop()
    
        if self._lock:
            self._lock.release()
    
    def _reset_handler(self, notification):
        """! @brief Reset notification handler.
        
        If the target is reset while the SWV reader is running, then the Target::trace_start()
        method is called to reinit trace output.
        """
        if self.is_alive():
            self._session.target.trace_start()




# pyOCD debugger
# Copyright (c) 2006-2013,2018-2021 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (Set, Tuple)

from .dap_access_api import DAPAccessIntf

class Command:
    DAP_INFO = 0x00
    DAP_LED = 0x01
    DAP_CONNECT = 0x02
    DAP_DISCONNECT = 0x03
    DAP_TRANSFER_CONFIGURE = 0x04
    DAP_TRANSFER = 0x05
    DAP_TRANSFER_BLOCK = 0x06
    DAP_TRANSFER_ABORT = 0x07
    DAP_WRITE_ABORT = 0x08
    DAP_DELAY = 0x09
    DAP_RESET_TARGET = 0x0a
    DAP_SWJ_PINS = 0x10
    DAP_SWJ_CLOCK = 0x11
    DAP_SWJ_SEQUENCE = 0x12
    DAP_SWD_CONFIGURE = 0x13
    DAP_JTAG_SEQUENCE = 0x14
    DAP_JTAG_CONFIGURE = 0x15
    DAP_JTAG_IDCODE = 0x16
    DAP_SWO_TRANSPORT = 0x17
    DAP_SWO_MODE = 0x18
    DAP_SWO_BAUDRATE = 0x19
    DAP_SWO_CONTROL = 0x1A
    DAP_SWO_STATUS = 0x1B
    DAP_SWO_DATA = 0x1C
    DAP_SWD_SEQUENCE = 0x1D
    DAP_SWO_EXTENDED_STATUS = 0x1E
    DAP_QUEUE_COMMANDS = 0x7E
    DAP_EXECUTE_COMMANDS = 0x7F
    DAP_VENDOR0 = 0x80 # Start of vendor-specific command IDs.

class Capabilities:
    SWD = 0x01
    JTAG = 0x02
    SWO_UART = 0x04
    SWO_MANCHESTER = 0x08
    ATOMIC_COMMANDS = 0x10
    DAP_SWD_SEQUENCE = 0x20

class Pin:
    NONE = 0x00 # Used to read current pin values without changing.
    SWCLK_TCK = (1 << 0)
    SWDIO_TMS = (1 << 1)
    TDI = (1 << 2)
    TDO = (1 << 3)
    nTRST = (1 << 5)
    nRESET = (1 << 7)

# Info IDs that return integer values.
INTEGER_INFOS = [
    DAPAccessIntf.ID.CAPABILITIES,
    DAPAccessIntf.ID.TEST_DOMAIN_TIMER,
    DAPAccessIntf.ID.SWO_BUFFER_SIZE,
    DAPAccessIntf.ID.MAX_PACKET_COUNT,
    DAPAccessIntf.ID.MAX_PACKET_SIZE
    ]

class CMSISDAPVersion:
    """! @brief Known CMSIS-DAP versions.

    The tuple fields are major, minor, patch. Generally, patch release versions are excluded from this
    list, unless there is a specific reason to know about a particular patch release.
    """
    V1_0_0 = (1, 0, 0)
    V1_1_0 = (1, 1, 0)
    V1_2_0 = (1, 2, 0)
    V1_3_0 = (1, 3, 0)
    V2_0_0 = (2, 0, 0)
    V2_1_0 = (2, 1, 0)

    @classmethod
    def minor_versions(cls) -> Set[Tuple[int, int]]:
        """@brief Returns a set of minor version tuples."""
        return {
            v[:2] for k, v in cls.__dict__.items()
            if k.startswith('V')
            }

DAP_DEFAULT_PORT = 0
DAP_SWD_PORT = 1
DAP_JTAG_PORT = 2

DAP_LED_CONNECT = 0
DAP_LED_RUNNING = 1

# Masks for DAP_SWD_SEQUENCE command.
class DAPSWDSequence:
    MODE_MASK = 0x80 # bit [7]: 0=output, 1=input
    CYCLES_MASK = 0x1f # bits [5:0]: number of TCK cycles

# Options for DAP_SWO_TRANSPORT command.
class DAPSWOTransport:
    NONE = 0
    DAP_SWO_DATA = 1
    DAP_SWO_EP = 2

# SWO mode options.
class DAPSWOMode:
    OFF = 0
    UART = 1
    MANCHESTER = 2

# SWO control actions.
class DAPSWOControl:
    STOP = 0
    START = 1

# SWO status masks.
class DAPSWOStatus:
    CAPTURE = 0x01
    ERROR = 0x40
    OVERRUN = 0x80

DAP_OK = 0
DAP_ERROR = 0xff

class DAPTransferResponse:
    """! Responses to DAP_Transfer and DAP_TransferBlock"""
    ACK_MASK = 0x07 # Bits [2:0]
    PROTOCOL_ERROR_MASK = 0x08 # Bit [3]
    VALUE_MISMATCH_MASK = 0x08 # Bit [4]
    
    # Values for ACK bitfield.
    ACK_OK = 1
    ACK_WAIT = 2
    ACK_FAULT = 4
    ACK_NO_ACK = 7

class CMSISDAPProtocol(object):
    """! @brief This class implements the CMSIS-DAP wire protocol."""

    def __init__(self, interface):
        self.interface = interface

    def dap_info(self, id_):
        """! @brief Sends the DAP_Info command to read info from the CMSIS-DAP probe.
        @param self This object.
        @param id_ One of the @ref pyocd.probe.pydapaccess.dap_access_api.DAPAcessIntf.ID "DAPAcessIntf.ID" constants.
        @return The `id_` parameter determines the return value data type. For those IDs defined as integer values
            in the DAP_Info command documentation, an int is returned. Otherwise either a string is returned, or
            None if the returned info value length is 0.
        @exception DAPAccessIntf.DeviceError Raised under these conditions:
            - Response packet is for a different command.
            - An int-type info was requested, but the returned value length is not 1, 2, or 4.
            - A string-type info was requested, but the returned value length is greater than the response packet size minus response header and terminating null byte on the string.
        """
        assert type(id_) is DAPAccessIntf.ID
            
        cmd = []
        cmd.append(Command.DAP_INFO)
        cmd.append(id_.value)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_INFO:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_INFO")

        resp_len = resp[1]

        # Integer values
        if id_ in INTEGER_INFOS:
            if resp_len == 1:
                return resp[2]
            elif resp_len == 2:
                return (resp[3] << 8) | resp[2]
            elif resp_len == 4:
                return (resp[5] << 24) | (resp[4] << 16) | (resp[3] << 8) | resp[2]
            else:
                raise DAPAccessIntf.DeviceError("invalid DAP_INFO response length for %s" % id_.name)

        # String values. They are sent as C strings with a terminating null char, so we strip it out.
        if resp_len == 0:
            return None
        if resp_len > (len(resp) - 2):
            raise DAPAccessIntf.DeviceError("invalid DAP_INFO response length for %s" % id_.name)
        return bytearray(resp[2:2 + resp_len - 1]).decode('utf-8', 'replace')

    def set_led(self, type, enabled):
        cmd = []
        cmd.append(Command.DAP_LED)
        cmd.append(type)
        cmd.append(int(enabled))
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_LED:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_LED")

        if resp[1] != 0:
            # Second response byte must be 0
            raise DAPAccessIntf.CommandError("DAP_LED failed")

        return resp[1]

    def connect(self, mode=DAP_DEFAULT_PORT):
        cmd = []
        cmd.append(Command.DAP_CONNECT)
        cmd.append(mode)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_CONNECT:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_CONNECT")

        if resp[1] == 0:
            # DAP connect failed
            raise DAPAccessIntf.CommandError("DAP_CONNECT failed")

        return resp[1]

    def disconnect(self):
        cmd = []
        cmd.append(Command.DAP_DISCONNECT)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_DISCONNECT:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_DISCONNECT")

        if resp[1] != DAP_OK:
            # DAP Disconnect failed
            raise DAPAccessIntf.CommandError("DAP_DISCONNECT failed")

        return resp[1]

    def write_abort(self, data, dap_index=0):
        cmd = []
        cmd.append(Command.DAP_WRITE_ABORT)
        cmd.append(dap_index)
        cmd.append((data >> 0) & 0xff)
        cmd.append((data >> 8) & 0xff)
        cmd.append((data >> 16) & 0xff)
        cmd.append((data >> 24) & 0xff)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_WRITE_ABORT:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_WRITE_ABORT")

        if resp[1] != DAP_OK:
            # DAP Write Abort failed
            raise DAPAccessIntf.CommandError("DAP_WRITE_ABORT failed")

        return True

    def reset_target(self):
        cmd = []
        cmd.append(Command.DAP_RESET_TARGET)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_RESET_TARGET:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_RESET_TARGET")

        if resp[1] != DAP_OK:
            # DAP Reset target failed
            raise DAPAccessIntf.CommandError("DAP_RESET_TARGET failed")

        return resp[1]

    def transfer_configure(self, idle_cycles=0x02, wait_retry=0x0050, match_retry=0x0000):
        cmd = []
        cmd.append(Command.DAP_TRANSFER_CONFIGURE)
        cmd.append(idle_cycles)
        cmd.append(wait_retry & 0xff)
        cmd.append(wait_retry >> 8)
        cmd.append(match_retry & 0xff)
        cmd.append(match_retry >> 8)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_TRANSFER_CONFIGURE:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_TRANSFER_CONFIGURE")

        if resp[1] != DAP_OK:
            # DAP Transfer Configure failed
            raise DAPAccessIntf.CommandError("DAP_TRANSFER_CONFIGURE failed")

        return resp[1]


    def set_swj_clock(self, clock=1000000):
        cmd = []
        cmd.append(Command.DAP_SWJ_CLOCK)
        cmd.append(clock & 0xff)
        cmd.append((clock >> 8) & 0xff)
        cmd.append((clock >> 16) & 0xff)
        cmd.append((clock >> 24) & 0xff)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_SWJ_CLOCK:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWJ_CLOCK")

        if resp[1] != DAP_OK:
            # DAP SWJ Clock failed
            raise DAPAccessIntf.CommandError("DAP_SWJ_CLOCK failed")

        return resp[1]

    def set_swj_pins(self, output, pins, wait=0):
        cmd = []
        cmd.append(Command.DAP_SWJ_PINS)
        cmd.append(output & 0xff)
        cmd.append(pins)
        cmd.append(wait & 0xff)
        cmd.append((wait >> 8) & 0xff)
        cmd.append((wait >> 16) & 0xff)
        cmd.append((wait >> 24) & 0xff)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_SWJ_PINS:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWJ_PINS")

        return resp[1]

    def swd_configure(self, turnaround=1, always_send_data_phase=False):
        assert 1 <= turnaround <= 4
        conf = (turnaround - 1) | (int(always_send_data_phase) << 2)
    
        cmd = []
        cmd.append(Command.DAP_SWD_CONFIGURE)
        cmd.append(conf)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_SWD_CONFIGURE:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWD_CONFIGURE")

        if resp[1] != DAP_OK:
            # DAP SWD Configure failed
            raise DAPAccessIntf.CommandError("DAP_SWD_CONFIGURE failed")

        return resp[1]

    def swd_sequence(self, sequences):
        """! @brief Send the DAP_SWD_Sequence command.
        
        Each sequence in the _sequences_ parameter is a tuple with 1 or 2 members:
        - 0: int: number of TCK cycles from 1-64
        - 1: int: the SWDIO bit values to transfer. The presence of this tuple member indicates the sequence is
            an output sequence; the absence means that the specified number of TCK cycles of SWDIO data will be
            read and returned.
        
        The DAP_SWD_Sequence command expects this data for each sequence:
        - 0: sequence info byte
            - bit [7]: mode, 0=output, 1=input
            - bit [6]: reserved
            - bits [5:0]: number of TCK cycles from 1-64, with 64 encoded as 0
        - 1: (only present if element 0 bit 7 == 0, for output mode) bytes of data to send, one bit per TCK cycle,
            transmitted LSB first.
        
        @param self
        @param sequences A sequence of sequence description tuples as described above.
        
        @return A 2-tuple of the response status, and a sequence of bytes objects, one for each input
            sequence. The length of the bytes object is (<TCK-count> + 7) / 8. Bits are in LSB first order.
        """
        cmd = [Command.DAP_SWD_SEQUENCE]
        cmd.append(len(sequences))
        for seq in sequences:
            # Construct the control byte.
            tck_count = seq[0]
            assert 1 <= tck_count <= 64, "SWD sequence TCK count is out of range (must be 1-64)"
            is_output = len(seq) == 2
            info = (0x00 if is_output else 0x80) | (0 if (tck_count == 64) else tck_count)
            cmd.append(info)
            
            # Append SWDIO output data.
            if is_output:
                bits = seq[1]
                for i in range((tck_count + 7) // 8):
                    cmd.append(bits & 0xff)
                    bits >>= 8
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_SWD_SEQUENCE:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWD_SEQUENCE")

        if resp[1] != DAP_OK:
            # DAP SWD Configure failed
            raise DAPAccessIntf.CommandError("DAP_SWD_SEQUENCE failed")

        result = []
        offset = 2
        for seq in sequences:
            # Only handle input sequences. Convert the bytes into an int, LSB first.
            if len(seq) == 1:
                tck_count = seq[0]
                byte_count = (tck_count + 7) // 8
                bits = 0
                for i in range(byte_count):
                    bits |= resp[offset + i] << (8 * i)
                offset += byte_count
                result.append(bits)

        return resp[1], result

    def swj_sequence(self, length, bits):
        assert 0 <= length <= 256
        cmd = []
        cmd.append(Command.DAP_SWJ_SEQUENCE)
        cmd.append(0 if (length == 256) else length)
        for i in range((length + 7) // 8):
            cmd.append(bits & 0xff)
            bits >>= 8
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_SWJ_SEQUENCE:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWJ_SEQUENCE")

        if resp[1] != DAP_OK:
            # DAP SWJ Sequence failed
            raise DAPAccessIntf.CommandError("DAP_SWJ_SEQUENCE failed")

        return resp[1]

    def jtag_sequence(self, cycles, tms, read_tdo, tdi):
        assert 0 <= cycles <= 64
        info = (((0 if (cycles == 64) else cycles) & 0x3f)
                | ((tms & 1) << 6)
                | (int(read_tdo) << 7))
        
        cmd = []
        cmd.append(Command.DAP_JTAG_SEQUENCE)
        cmd.append(1)
        cmd.append(info)
        for i in range((cycles + 7) // 8):
            cmd.append(tdi & 0xff)
            tdi >>= 8
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_JTAG_SEQUENCE:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("exected DAP_JTAG_SEQUENCE")

        if resp[1] != DAP_OK:
            # DAP JTAG Sequence failed
            raise DAPAccessIntf.CommandError("DAP_JTAG_SEQUENCE failed")

        return resp[2]

    def jtag_configure(self, devices_irlen=None):
        # Default to a single device with an IRLEN of 4.
        if devices_irlen is None:
            devices_irlen = [4]
        
        cmd = []
        cmd.append(Command.DAP_JTAG_CONFIGURE)
        cmd.append(len(devices_irlen))
        for irlen in devices_irlen:
            cmd.append(irlen)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_JTAG_CONFIGURE:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_JTAG_CONFIGURE")

        if resp[1] != DAP_OK:
            # DAP JTAG Configure failed
            raise DAPAccessIntf.CommandError("DAP_JTAG_CONFIGURE failed")

        return resp[2:]

    def jtag_id_code(self, index=0):
        cmd = []
        cmd.append(Command.DAP_JTAG_IDCODE)
        cmd.append(index)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_JTAG_IDCODE:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_JTAG_IDCODE")

        if resp[1] != DAP_OK:
            # Operation failed
            raise DAPAccessIntf.CommandError("DAP_JTAG_IDCODE failed")

        return  (resp[2] << 0) | \
                (resp[3] << 8) | \
                (resp[4] << 16) | \
                (resp[5] << 24)

    def swo_transport(self, transport):
        cmd = []
        cmd.append(Command.DAP_SWO_TRANSPORT)
        cmd.append(transport)
        self.interface.write(cmd)

        resp = self.interface.read()

        if resp[0] != Command.DAP_SWO_TRANSPORT:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWO_TRANSPORT")

        if resp[1] != DAP_OK:
            # Operation failed
            raise DAPAccessIntf.CommandError("DAP_SWO_TRANSPORT failed")

        return resp[1]

    def swo_mode(self, mode):
        cmd = []
        cmd.append(Command.DAP_SWO_MODE)
        cmd.append(mode)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_SWO_MODE:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWO_MODE")

        if resp[1] != DAP_OK:
            # Operation failed
            raise DAPAccessIntf.CommandError("DAP_SWO_MODE failed")

        return resp[1]

    def swo_baudrate(self, baudrate):
        cmd = []
        cmd.append(Command.DAP_SWO_BAUDRATE)
        cmd.append(baudrate & 0xff)
        cmd.append((baudrate >> 8) & 0xff)
        cmd.append((baudrate >> 16) & 0xff)
        cmd.append((baudrate >> 24) & 0xff)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_SWO_BAUDRATE:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWO_BAUDRATE")

        return  (resp[1] << 0) | \
                (resp[2] << 8) | \
                (resp[3] << 16) | \
                (resp[4] << 24)

    def swo_control(self, action):
        cmd = []
        cmd.append(Command.DAP_SWO_CONTROL)
        cmd.append(action)
        self.interface.write(cmd)

        resp = self.interface.read()

        if resp[0] != Command.DAP_SWO_CONTROL:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWO_CONTROL")

        if resp[1] != DAP_OK:
            # Operation failed
            raise DAPAccessIntf.CommandError("DAP_SWO_CONTROL failed")

        return resp[1]

    def swo_status(self):
        cmd = []
        cmd.append(Command.DAP_SWO_STATUS)
        self.interface.write(cmd)

        resp = self.interface.read()
        if resp[0] != Command.DAP_SWO_STATUS:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWO_STATUS")

        return (resp[1],
                    (resp[2] << 0) | \
                    (resp[3] << 8) | \
                    (resp[4] << 16) | \
                    (resp[5] << 24)
                )

    def swo_data(self, count):
        cmd = []
        cmd.append(Command.DAP_SWO_DATA)

        # Account for protocol overhead when setting the count
        count = min(self.interface.get_packet_size() - 4, count)
        cmd.append(count & 0xff)
        cmd.append((count >> 8) & 0xff)

        self.interface.write(cmd)
        resp = self.interface.read()

        if resp[0] != Command.DAP_SWO_DATA:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_SWO_DATA")

        status = resp[1]
        count = (resp[2] << 0) | \
                    (resp[3] << 8)
        if count > 0:
            data = resp[4:]
        else:
            data = []
        return (status, count, data)

    def vendor(self, index, data):
        cmd = []
        cmd.append(Command.DAP_VENDOR0 + index)
        cmd.extend(data)
        self.interface.write(cmd)

        resp = self.interface.read()

        if resp[0] != Command.DAP_VENDOR0 + index:
            # Response is to a different command
            raise DAPAccessIntf.DeviceError("expected DAP_VENDOR%i" % index)

        return resp[1:]



# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pkg_resources

from .. import __version__
from ..core.session import Session
from ..core.helpers import ConnectHelper
from ..core import options
from ..target import TARGET
from ..target.builtin import BUILTIN_TARGETS
from ..board.board_ids import BOARD_ID_TO_INFO
from ..target.pack import pack_target

class ListGenerator(object):
    @staticmethod
    def list_probes():
        """! @brief Generate dictionary with info about the connected debug probes.
        
        Output version history:
        - 1.0, initial version
        """
        status = 0
        error = ""
        try:
            all_mbeds = ConnectHelper.get_sessions_for_all_connected_probes(blocking=False)
        except Exception as e:
            all_mbeds = []
            status = 1
            error = str(e)

        boards = []
        obj = {
            'pyocd_version' : __version__,
            'version' : { 'major' : 1, 'minor' : 0 },
            'status' : status,
            'boards' : boards,
            }

        if status != 0:
            obj['error'] = error

        for mbed in all_mbeds:
            d = {
                'unique_id' : mbed.probe.unique_id,
                'info' : mbed.board.description,
                'board_name' : mbed.board.name,
                'target' : mbed.board.target_type,
                'vendor_name' : mbed.probe.vendor_name,
                'product_name' : mbed.probe.product_name,
                }
            boards.append(d)

        return obj

    @staticmethod
    def list_boards(name_filter=None):
        """! @brief Generate dictionary with info about supported boards.
        
        Output version history:
        - 1.0, initial version
        - 1.1, added is_target_builtin and is_target_supported keys
        """
        # Lowercase name and vendor arguments for case-insensitive comparison.
        if name_filter is not None:
            name_filter = name_filter.lower()

        boards = []
        obj = {
            'pyocd_version' : __version__,
            'version' : { 'major' : 1, 'minor' : 1 },
            'status' : 0,
            'boards' : boards
            }

        # Lowercase target names for comparison
        managed_targets = [dev.part_number.lower() for dev in pack_target.ManagedPacks.get_installed_targets()]
        builtin_target_names = [target_name.lower() for target_name in BUILTIN_TARGETS]
        target_names = [target_name.lower() for target_name in TARGET]

        for board_id, info in BOARD_ID_TO_INFO.items():
            # Filter by name.
            if name_filter and name_filter not in info.name.lower():
                continue
            d = {
                'id' : board_id,
                'name' : info.name,
                'target': info.target,
                'binary' : info.binary,
                'is_target_builtin': (info.target.lower() in builtin_target_names),
                'is_target_supported': (info.target.lower() in target_names or info.target in managed_targets)
                }
            boards.append(d)

        return obj

    @staticmethod
    def list_targets(name_filter=None, vendor_filter=None, source_filter=None):
        """! @brief Generate dictionary with info about all supported targets.
        
        Output version history:
        - 1.0, initial version
        - 1.1, added part_families
        - 1.2, added source
        """
        # Lowercase name and vendor arguments for case-insensitive comparison.
        if name_filter is not None:
            name_filter = name_filter.lower()
        if vendor_filter is not None:
            vendor_filter = vendor_filter.lower()

        targets = []
        obj = {
            'pyocd_version' : __version__,
            'version' : { 'major' : 1, 'minor' : 2 },
            'status' : 0,
            'targets' : targets
            }

        for name in TARGET.keys():
            # Filter by name.
            if name_filter and name_filter not in name.lower():
                continue
            
            s = Session(None) # Create empty session
            t = TARGET[name](s)
            
            # Filter by vendor.
            if vendor_filter and vendor_filter not in t.vendor.lower():
                continue
            
            # Filter by source.
            source = 'pack' if hasattr(t, '_pack_device') else 'builtin'
            if source_filter and source_filter != source:
                continue
            
            d = {
                'name' : name,
                'vendor' : t.vendor,
                'part_families' : t.part_families,
                'part_number' : t.part_number,
                'source': source,
                }
            if t._svd_location is not None:
                svdPath = t._svd_location.filename
                if isinstance(svdPath, str) and os.path.exists(svdPath):
                    d['svd_path'] = svdPath
            targets.append(d)
        
        if not source_filter or source_filter == 'pack':
            # Add targets from cmsis-pack-manager cache.
            for dev in pack_target.ManagedPacks.get_installed_targets():
                try:
                    # Filter by name.
                    if name_filter and name_filter not in dev.part_number.lower():
                        continue
                    # Filter by vendor.
                    if vendor_filter and vendor_filter not in dev.vendor.lower():
                        continue
                    targets.append({
                        'name' : dev.part_number.lower(),
                        'part_families' : dev.families,
                        'part_number' : dev.part_number,
                        'vendor' : dev.vendor,
                        'source' : 'pack',
                        })
                except KeyError:
                    pass

        return obj
        
    @staticmethod
    def list_plugins():
        """! @brief Generate dictionary with lists of available plugins.
        
        Output version history:
        - 1.0, initial version with debug probe and RTOS plugins
        """
        from ..probe.aggregator import PROBE_CLASSES
        from ..rtos import RTOS
        plugin_groups = [
                'pyocd.probe',
                'pyocd.rtos',
                ]
        plugin_groups_list = []
        obj = {
            'pyocd_version': __version__,
            'version': { 'major': 1, 'minor': 0 },
            'status': 0,
            'plugins': plugin_groups_list,
            }
        
        # Add plugins info
        for group_name in plugin_groups:
            plugin_list = []
            group_info = {
                'plugin_type': group_name,
                'plugins': plugin_list,
                }
            
            for entry_point in pkg_resources.iter_entry_points(group_name):
                klass = entry_point.load()
                plugin = klass()
                info = {
                    'name': plugin.name,
                    'version': plugin.version,
                    'description': plugin.description,
                    'classname': klass.__name__,
                    }
                plugin_list.append(info)
            plugin_groups_list.append(group_info)
        
        return obj
        
    @staticmethod
    def list_features():
        """! @brief Generate dictionary with info about supported features and options.
        
        Output version history:
        - 1.1, added 'plugins' feature
        - 1.0, initial version
        """
        options_list = []
        plugins_list = []
        obj = {
            'pyocd_version' : __version__,
            'version' : { 'major' : 1, 'minor' : 1 },
            'status' : 0,
            'features' : [
                    {
                        'name': 'plugins',
                        'plugins': plugins_list,
                    },
                ],
            'options' : options_list,
            }
        
        # Add plugins
        plugins = ListGenerator.list_plugins()
        plugins_list.extend(plugins['plugins'])
        
        # Add options
        for option_name in options.OPTIONS_INFO.keys():
            info = options.OPTIONS_INFO[option_name]
            option_dict = {
                        'name' : option_name,
                        'default' : info.default,
                        'description' : info.help,
                        }
            try:
                types_list = []
                for t in info.type:
                    types_list.append(t.__name__)
            except TypeError:
                types_list = [info.type.__name__]
            option_dict['type'] = types_list
            options_list.append(option_dict)
        
        return obj



# pyOCD debugger
# Copyright (c) 2020 Wagner Sartori Junior
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile

class DBGMCU:
    CR = 0xE0042004
    CR_VALUE = 0x7 # DBG_STANDBY | DBG_STOP | DBG_SLEEP

    APB1FZR1 = 0xE0042008
    APB1FZR1_VALUE = 0b10000010111000000001110000110001

    APB1FZR2 = 0xE004200C
    APB1FZR2_VALUE = 0b00000000000000000000000000100000

    APB2FZR = 0xE0042010
    APB2FZR_VALUE = 0b00000000000000110000100000000000

FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x8f4ff3bf, 0x48584770, 0x49586800, 0x0d000500, 0xd0001840, 0x47702001, 0x6a004855, 0x0fc00280,
    0xb5004770, 0xf7ff4602, 0x2801ffee, 0xf7ffd108, 0x2801fff3, 0x484fd104, 0xd3014282, 0xbd002001,
    0xbd002000, 0x4602b500, 0xffddf7ff, 0xd0022801, 0x0d8002d0, 0x4948bd00, 0x40080ad0, 0xd5f90391,
    0x300130ff, 0x4842bd00, 0x60814944, 0x60814944, 0x60012100, 0x61014943, 0x03c06a00, 0x4843d406,
    0x60014941, 0x60412106, 0x60814941, 0x47702000, 0x49372001, 0x614807c0, 0x47702000, 0x47702001,
    0x49384833, 0x13c16101, 0x69416141, 0x04122201, 0x61414311, 0x4a354937, 0x6011e000, 0x03db6903,
    0x2100d4fb, 0x46086141, 0xb5104770, 0xf7ff4604, 0x4603ffa8, 0xf7ff4620, 0x4925ffb5, 0x610c4c29,
    0x02d800c2, 0x43021c92, 0x6948614a, 0x04122201, 0x61484310, 0x8f4ff3bf, 0x4a244826, 0x6010e000,
    0x03db690b, 0x2000d4fb, 0x69086148, 0xd0014020, 0x2001610c, 0xb5f0bd10, 0x4d151dc9, 0x4f1908c9,
    0x612f00c9, 0x616b2300, 0xe0184c1a, 0x616b2301, 0x60036813, 0x60436853, 0x8f4ff3bf, 0xe0004b13,
    0x692e601c, 0xd4fb03f6, 0x616b2300, 0x423b692b, 0x612fd002, 0xbdf02001, 0x39083008, 0x29003208,
    0x2000d1e4, 0x0000bdf0, 0xe0042000, 0xfffffbcb, 0x40022000, 0x08020000, 0x000003bf, 0x45670123,
    0xcdef89ab, 0x0000c3fa, 0x00005555, 0x40003000, 0x00000fff, 0x0000aaaa, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000087,
    'pc_unInit': 0x200000b1,
    'pc_program_page': 0x20000137,
    'pc_erase_sector': 0x200000eb,
    'pc_eraseAll': 0x200000c1,

    'static_base' : 0x20000000 + 0x00000020 + 0x00000198,
    'begin_stack' : 0x20000400,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x400,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20001000, 0x20001400],   # Enable double buffering
    'min_program_length' : 0x400,

    # Flash information
    'flash_start': 0x8000000,
    'flash_size': 0x40000,
    'sector_sizes': (
        (0x0, 0x800),
    )
}

class STM32L432xC(CoreSightTarget):
        
    VENDOR = "STMicroelectronics"

    MEMORY_MAP = MemoryMap(
        FlashRegion(name='flash', start=0x08000000, length=0x40000,
                        sector_size=0x800,
                        page_size=0x400,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(name='sram1',   start=0x20000000, length=0xC000),
        RamRegion(name='sram2',   start=0x10000000, length=0x4000)
        )

    def __init__(self, session):
        super(STM32L432xC, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("STM32L4x2.svd")

    def post_connect_hook(self):
        self.write32(DBGMCU.CR, DBGMCU.CR_VALUE)
        self.write32(DBGMCU.APB1FZR1, DBGMCU.APB1FZR1_VALUE)
        self.write32(DBGMCU.APB1FZR2, DBGMCU.APB1FZR2_VALUE)
        self.write32(DBGMCU.APB2FZR, DBGMCU.APB2FZR_VALUE)




# pyOCD debugger
# Copyright (c) 2006-2015,2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function
import sys
from struct import unpack

INPUT_FILENAME = sys.argv[1]
OUTPUT_FILENAME = sys.argv[2]

with open(INPUT_FILENAME, "rb") as f:
    data = f.read()

words = len(data) // 4
if len(data) % 4 != 0:
    print("Warning: input length not word aligned")
print("Data length %i" % len(data))
data = unpack("<%iL" % words, data)

str = "analyzer = (\n    "
count = 0
for val in data:
    if count % 8 == 7:
        str += "0x{:08x},\n    ".format(val)
    else:
        str += "0x{:08x}, ".format(val)
    count += 1
str += "\n    )"
data = str

with open(OUTPUT_FILENAME, "w") as f:
    f.write(data)



# pyOCD debugger
# Copyright (c) 2006-2013,2018-2021 Arm Limited
# Copyright (c) 2020 Koji Kitayama
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import logging
import collections
import threading
from typing import (Optional, Tuple)

from .dap_settings import DAPSettings
from .dap_access_api import DAPAccessIntf
from .cmsis_dap_core import CMSISDAPProtocol
from .interface import (INTERFACE, USB_BACKEND, USB_BACKEND_V2)
from .cmsis_dap_core import (
    Command,
    Pin,
    Capabilities,
    DAPSWOTransport,
    DAPSWOMode,
    DAPSWOControl,
    DAPTransferResponse,
    CMSISDAPVersion,
    )
from ...core import session
from ...utility.concurrency import locked

VersionTuple = Tuple[int, int, int]

# CMSIS-DAP values
AP_ACC = 1 << 0
DP_ACC = 0 << 0
READ = 1 << 1
WRITE = 0 << 1
VALUE_MATCH = 1 << 4
MATCH_MASK = 1 << 5

# SWO statuses.
class SWOStatus:
    DISABLED = 1
    CONFIGURED = 2
    RUNNING = 3
    ERROR = 4

LOG = logging.getLogger(__name__)

TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

def _get_interfaces():
    """! @brief Get the connected USB devices"""
    # Get CMSIS-DAPv1 interfaces.
    v1_interfaces = INTERFACE[USB_BACKEND].get_all_connected_interfaces()
    
    # Get CMSIS-DAPv2 interfaces.
    v2_interfaces = INTERFACE[USB_BACKEND_V2].get_all_connected_interfaces()
    
    # Prefer v2 over v1 if a device provides both.
    devices_in_both = [v1 for v1 in v1_interfaces for v2 in v2_interfaces
                        if _get_unique_id(v1) == _get_unique_id(v2)]
    for dev in devices_in_both:
        v1_interfaces.remove(dev)
        
    # Return the combined list.
    return v1_interfaces + v2_interfaces


def _get_unique_id(interface):
    """! @brief Get the unique id from an interface"""
    return interface.get_serial_number()


class _Transfer(object):
    """! @brief A wrapper object representing a command invoked by the layer above.

    The transfer class contains a logical register read or a block
    of reads to the same register.
    """

    def __init__(self, daplink, dap_index, transfer_count,
                 transfer_request, transfer_data):
        # Writes should not need a transfer object
        # since they don't have any response data
        assert isinstance(dap_index, int)
        assert isinstance(transfer_count, int)
        assert isinstance(transfer_request, int)
        assert transfer_request & READ
        self.daplink = daplink
        self.dap_index = dap_index
        self.transfer_count = transfer_count
        self.transfer_request = transfer_request
        self.transfer_data = transfer_data
        self._size_bytes = 0
        if transfer_request & READ:
            self._size_bytes = transfer_count * 4
        self._result = None
        self._error = None

    def get_data_size(self):
        """! @brief Get the size in bytes of the return value of this transfer
        """
        return self._size_bytes

    def add_response(self, data):
        """! @brief Add data read from the remote device to this object.

        The size of data added must match exactly the size
        that get_data_size returns.
        """
        assert len(data) == self._size_bytes
        result = []
        for i in range(0, self._size_bytes, 4):
            word = ((data[0 + i] << 0) | (data[1 + i] << 8) |
                    (data[2 + i] << 16) | (data[3 + i] << 24))
            result.append(word)
        self._result = result

    def add_error(self, error):
        """! @brief Attach an exception to this transfer rather than data.
        """
        assert isinstance(error, Exception)
        self._error = error

    def get_result(self):
        """! @brief Get the result of this transfer.
        """
        while self._result is None:
            if len(self.daplink._commands_to_read) > 0:
                self.daplink._read_packet()
            else:
                assert not self.daplink._crnt_cmd.get_empty()
                self.daplink.flush()

        if self._error is not None:
            # Pylint is confused and thinks self._error is None
            # since that is what it is initialized to.
            # Suppress warnings for this.
            # pylint: disable=raising-bad-type
            raise self._error

        assert self._result is not None
        return self._result

class _Command(object):
    """! @brief Wrapper object representing a command sent to the layer below (ex. USB).

    This class wraps the physical commands DAP_Transfer and DAP_TransferBlock
    to provide a uniform way to build the command to most efficiently transfer
    the data supplied.  Register reads and writes individually or in blocks
    are added to a command object until it is full.  Once full, this class
    decides if it is more efficient to use DAP_Transfer or DAP_TransferBlock.
    The payload to send over the layer below is constructed with
    encode_data.  The response to the command is decoded with decode_data.
    """

    def __init__(self, size):
        self._size = size
        self._read_count = 0
        self._write_count = 0
        self._block_allowed = True
        self._block_request = None
        self._data = []
        self._dap_index = None
        self._data_encoded = False
        TRACE.debug("New _Command")

    def _get_free_words(self, blockAllowed, isRead):
        """! @brief Return the number of words free in the transmit packet
        """
        if blockAllowed:
            # DAP_TransferBlock request packet:
            #   BYTE | BYTE *****| SHORT**********| BYTE *************| WORD *********|
            # > 0x06 | DAP Index | Transfer Count | Transfer Request  | Transfer Data |
            #  ******|***********|****************|*******************|+++++++++++++++|
            send = self._size - 5 - 4 * self._write_count

            # DAP_TransferBlock response packet:
            #   BYTE | SHORT *********| BYTE *************| WORD *********|
            # < 0x06 | Transfer Count | Transfer Response | Transfer Data |
            #  ******|****************|*******************|+++++++++++++++|
            recv = self._size - 4 - 4 * self._read_count

            if isRead:
                return recv // 4
            else:
                return send // 4
        else:
            # DAP_Transfer request packet:
            #   BYTE | BYTE *****| BYTE **********| BYTE *************| WORD *********|
            # > 0x05 | DAP Index | Transfer Count | Transfer Request  | Transfer Data |
            #  ******|***********|****************|+++++++++++++++++++++++++++++++++++|
            send = self._size - 3 - 1 * self._read_count - 5 * self._write_count

            # DAP_Transfer response packet:
            #   BYTE | BYTE **********| BYTE *************| WORD *********|
            # < 0x05 | Transfer Count | Transfer Response | Transfer Data |
            #  ******|****************|*******************|+++++++++++++++|
            recv = self._size - 3 - 4 * self._read_count

            if isRead:
                # 1 request byte in request packet, 4 data bytes in response packet
                return min(send, recv // 4)
            else:
                # 1 request byte + 4 data bytes
                return send // 5

    def get_request_space(self, count, request, dap_index):
        assert self._data_encoded is False

        # Must create another command if the dap index is different.
        if self._dap_index is not None and dap_index != self._dap_index:
            return 0

        # Block transfers must use the same request.
        blockAllowed = self._block_allowed
        if self._block_request is not None and request != self._block_request:
            blockAllowed = False

        # Compute the portion of the request that will fit in this packet.
        is_read = request & READ
        free = self._get_free_words(blockAllowed, is_read)
        size = min(count, free)

        # Non-block transfers only have 1 byte for request count.
        if not blockAllowed:
            max_count = self._write_count + self._read_count + size
            delta = max_count - 255
            size = min(size - delta, size)
            TRACE.debug("get_request_space(%d, %02x:%s)[wc=%d, rc=%d, ba=%d->%d] -> (sz=%d, free=%d, delta=%d)" %
                    (count, request, 'r' if is_read else 'w', self._write_count, self._read_count, self._block_allowed, blockAllowed, size, free, delta))
        else:
            TRACE.debug("get_request_space(%d, %02x:%s)[wc=%d, rc=%d, ba=%d->%d] -> (sz=%d, free=%d)" %
                (count, request, 'r' if is_read else 'w', self._write_count, self._read_count, self._block_allowed, blockAllowed, size, free))

        # We can get a negative free count if the packet already contains more data than can be
        # sent by a DAP_Transfer command, but the new request forces DAP_Transfer. In this case,
        # just return 0 to force the DAP_Transfer_Block to be sent.
        return max(size, 0)

    def get_full(self):
        return (self._get_free_words(self._block_allowed, True) == 0) or \
            (self._get_free_words(self._block_allowed, False) == 0)

    def get_empty(self):
        """! @brief Return True if no transfers have been added to this packet
        """
        return len(self._data) == 0

    def add(self, count, request, data, dap_index):
        """! @brief Add a single or block register transfer operation to this command
        """
        assert self._data_encoded is False
        if self._dap_index is None:
            self._dap_index = dap_index
        assert self._dap_index == dap_index

        if self._block_request is None:
            self._block_request = request
        elif request != self._block_request:
            self._block_allowed = False
        assert not self._block_allowed or self._block_request == request

        if request & READ:
            self._read_count += count
        else:
            self._write_count += count
        self._data.append((count, request, data))

        TRACE.debug("add(%d, %02x:%s) -> [wc=%d, rc=%d, ba=%d]" %
                (count, request, 'r' if (request & READ) else 'w', self._write_count, self._read_count, self._block_allowed))

    def _encode_transfer_data(self):
        """! @brief Encode this command into a byte array that can be sent

        The data returned by this function is a bytearray in
        the format that of a DAP_Transfer CMSIS-DAP command.
        """
        assert self.get_empty() is False
        buf = bytearray(self._size)
        transfer_count = self._read_count + self._write_count
        pos = 0
        buf[pos] = Command.DAP_TRANSFER
        pos += 1
        buf[pos] = self._dap_index
        pos += 1
        buf[pos] = transfer_count
        pos += 1
        for count, request, write_list in self._data:
            assert write_list is None or len(write_list) <= count
            write_pos = 0
            for _ in range(count):
                buf[pos] = request
                pos += 1
                if not request & READ:
                    buf[pos] = (write_list[write_pos] >> (8 * 0)) & 0xff
                    pos += 1
                    buf[pos] = (write_list[write_pos] >> (8 * 1)) & 0xff
                    pos += 1
                    buf[pos] = (write_list[write_pos] >> (8 * 2)) & 0xff
                    pos += 1
                    buf[pos] = (write_list[write_pos] >> (8 * 3)) & 0xff
                    pos += 1
                    write_pos += 1
        return buf[:pos]

    def _check_response(self, response):
        """! @brief Check the response status byte from CMSIS-DAP transfer commands.

        The ACK bits [2:0] and the protocol error bit are checked. If any error is indicated,
        the appropriate exception is raised. An exception is also raised for unrecognised ACK
        values.
        
        @param self
        @param response The "Transfer Response" byte from a DAP_Transfer or DAP_TransferBlock
            command.
        
        @exception DAPAccessIntf.TransferFaultError Raised for the ACK_FAULT response.
        @exception DAPAccessIntf.TransferTimeoutError Raised for ACK_WAIT response.
        @exception DAPAccessIntf.TransferError Raised for other, less common errors, including No
            ACK, SWD protocol error, and an unknown ACK value. A descriptive error message is used
            to indicate which of these errors was the cause.
        """
        ack = response & DAPTransferResponse.ACK_MASK
        if ack != DAPTransferResponse.ACK_OK:
            if ack == DAPTransferResponse.ACK_FAULT:
                raise DAPAccessIntf.TransferFaultError()
            elif ack == DAPTransferResponse.ACK_WAIT:
                raise DAPAccessIntf.TransferTimeoutError()
            elif ack == DAPTransferResponse.ACK_NO_ACK:
                raise DAPAccessIntf.TransferError("No ACK received")
            else:
                raise DAPAccessIntf.TransferError("Unexpected ACK value (%d) returned by probe" % ack)
        elif (response & DAPTransferResponse.PROTOCOL_ERROR_MASK) != 0:
            raise DAPAccessIntf.TransferError("SWD protocol error")
    
    def _decode_transfer_data(self, data):
        """! @brief Take a byte array and extract the data from it

        Decode the response returned by a DAP_Transfer CMSIS-DAP command
        and return it as an array of bytes.
        """
        assert self.get_empty() is False
        if data[0] != Command.DAP_TRANSFER:
            raise ValueError('DAP_TRANSFER response error')

        # Check response and raise an exception on errors.
        self._check_response(data[2])

        # Check for count mismatch after checking for DAP_TRANSFER_FAULT
        # This allows TransferFaultError or TransferTimeoutError to get
        # thrown instead of TransferFaultError
        if data[1] != self._read_count + self._write_count:
            raise DAPAccessIntf.TransferError()

        return data[3:3 + 4 * self._read_count]

    def _encode_transfer_block_data(self):
        """! @brief Encode this command into a byte array that can be sent

        The data returned by this function is a bytearray in
        the format that of a DAP_TransferBlock CMSIS-DAP command.
        """
        assert self.get_empty() is False
        buf = bytearray(self._size)
        transfer_count = self._read_count + self._write_count
        assert not (self._read_count != 0 and self._write_count != 0)
        assert self._block_request is not None
        pos = 0
        buf[pos] = Command.DAP_TRANSFER_BLOCK
        pos += 1
        buf[pos] = self._dap_index
        pos += 1
        buf[pos] = transfer_count & 0xff
        pos += 1
        buf[pos] = (transfer_count >> 8) & 0xff
        pos += 1
        buf[pos] = self._block_request
        pos += 1
        for count, request, write_list in self._data:
            assert write_list is None or len(write_list) <= count
            assert request == self._block_request
            write_pos = 0
            if not request & READ:
                for _ in range(count):
                    buf[pos] = (write_list[write_pos] >> (8 * 0)) & 0xff
                    pos += 1
                    buf[pos] = (write_list[write_pos] >> (8 * 1)) & 0xff
                    pos += 1
                    buf[pos] = (write_list[write_pos] >> (8 * 2)) & 0xff
                    pos += 1
                    buf[pos] = (write_list[write_pos] >> (8 * 3)) & 0xff
                    pos += 1
                    write_pos += 1
        return buf[:pos]

    def _decode_transfer_block_data(self, data):
        """! @brief Take a byte array and extract the data from it

        Decode the response returned by a DAP_TransferBlock CMSIS-DAP command
        and return it as an array of bytes.
        """
        assert self.get_empty() is False
        if data[0] != Command.DAP_TRANSFER_BLOCK:
            raise ValueError('DAP_TRANSFER_BLOCK response error')

        # Check response and raise an exception on errors.
        self._check_response(data[3])

        # Check for count mismatch after checking for DAP_TRANSFER_FAULT
        # This allows TransferFaultError or TransferTimeoutError to get
        # thrown instead of TransferFaultError
        transfer_count = data[1] | (data[2] << 8)
        if transfer_count != self._read_count + self._write_count:
            raise DAPAccessIntf.TransferError()

        return data[4:4 + 4 * self._read_count]

    def encode_data(self):
        """! @brief Encode this command into a byte array that can be sent

        The actual command this is encoded into depends on the data
        that was added.
        """
        assert self.get_empty() is False
        self._data_encoded = True
        if self._block_allowed:
            data = self._encode_transfer_block_data()
        else:
            data = self._encode_transfer_data()
        return data

    def decode_data(self, data):
        """! @brief Decode the response data
        """
        assert self.get_empty() is False
        assert self._data_encoded is True
        if self._block_allowed:
            data = self._decode_transfer_block_data(data)
        else:
            data = self._decode_transfer_data(data)
        return data

class DAPAccessCMSISDAP(DAPAccessIntf):
    """! @brief An implementation of the DAPAccessIntf layer for DAPLink boards
    
    @internal
    All methods that use the CMSISDAPProtocol instance must be locked and must flush the command queue
    prior to using methods of that object. Otherwise the command responses may be processed out of order.
    """

    # ------------------------------------------- #
    #          Static Functions
    # ------------------------------------------- #
    @staticmethod
    def get_connected_devices():
        """! @brief Return an array of all mbed boards connected
        """
        all_daplinks = []
        all_interfaces = _get_interfaces()
        for interface in all_interfaces:
            try:
                new_daplink = DAPAccessCMSISDAP(None, interface=interface)
                all_daplinks.append(new_daplink)
            except DAPAccessIntf.TransferError:
                LOG.error('Failed to get unique id', exc_info=session.Session.get_current().log_tracebacks)
        return all_daplinks

    @staticmethod
    def get_device(device_id):
        assert isinstance(device_id, str)
        iface = DAPAccessCMSISDAP._lookup_interface_for_unique_id(device_id)
        if iface is not None:
            return DAPAccessCMSISDAP(device_id, iface)
        else:
            return None

    @staticmethod
    def set_args(arg_list):
        # Example: arg_list =['limit_packets=True']
        arg_pattern = re.compile("([^=]+)=(.*)")
        if arg_list:
            for arg in arg_list:
                match = arg_pattern.match(arg)
                # check if arguments have correct format
                if match:
                    attr = match.group(1)
                    if hasattr(DAPSettings, attr):
                        val = match.group(2)
                        # convert string to int or bool
                        if val.isdigit():
                            val = int(val)
                        elif val == "True":
                            val = True
                        elif val == "False":
                            val = False
                        setattr(DAPSettings, attr, val)

    @staticmethod
    def _lookup_interface_for_unique_id(unique_id):
        result_interface = None
        all_interfaces = _get_interfaces()
        for interface in all_interfaces:
            try:
                if _get_unique_id(interface) == unique_id:
                    # This assert could indicate that two boards
                    # had the same ID
                    assert result_interface is None, "More than one probes with ID {}".format(unique_id)
                    result_interface = interface
            except Exception:
                LOG.error('Failed to get unique id for open', exc_info=session.Session.get_current().log_tracebacks)
        return result_interface

    # ------------------------------------------- #
    #          CMSIS-DAP and Other Functions
    # ------------------------------------------- #
    def __init__(self, unique_id, interface=None):
        assert isinstance(unique_id, str) or (unique_id is None and interface is not None)
        super(DAPAccessCMSISDAP, self).__init__()

        # Search for a matching interface if one wasn't provided.
        if interface is None:
            interface = DAPAccessCMSISDAP._lookup_interface_for_unique_id(unique_id)
            if interface is None:
                raise self.DeviceError("no device with ID %s" % unique_id)

        if interface is not None:
            self._unique_id = _get_unique_id(interface)
            self._vendor_name = interface.vendor_name
            self._product_name = interface.product_name
            self._vidpid = (interface.vid, interface.pid)
        else:
            # Set default values for an unknown interface.
            self._unique_id = unique_id
            self._vendor_name = ""
            self._product_name = ""
            self._vidpid = (0, 0)
            
        self._lock = threading.RLock()
        self._interface = interface
        self._deferred_transfer = False
        self._protocol = None  # TODO, c1728p9 remove when no longer needed
        self._packet_count = None
        self._frequency = 1000000  # 1MHz default clock
        self._dap_port = None
        self._transfer_list = None
        self._crnt_cmd = None
        self._packet_size = None
        self._commands_to_read = None
        self._command_response_buf = None
        self._swo_status = None
        self._cmsis_dap_version: VersionTuple = CMSISDAPVersion.V1_0_0
        self._fw_version: Optional[str] = None

    @property
    def protocol_version(self) -> VersionTuple:
        """! @brief Tuple of CMSIS-DAP protocol version.
        @return 3-tuple consisting of (major, minor, micro) version of the CMSIS-DAP protocol implemented
            by the debug probe.
        """
        return self._cmsis_dap_version
    
    @property
    def firmware_version(self) -> Optional[str]:
        """! @brief A string of the product firmware version, or None.
        
        Only probes supporting CMSIS-DAP protocol v2.1 or later can return their firmware version.
        """
        return self._fw_version

    @property
    def vendor_name(self):
        return self._vendor_name

    @property
    def product_name(self):
        return self._product_name
    
    @property
    def vidpid(self):
        """! @brief A tuple of USB VID and PID, in that order."""
        return self._vidpid

    @property
    def has_swd_sequence(self):
        return self._cmsis_dap_version >= CMSISDAPVersion.V1_2_0
    
    def lock(self):
        """! @brief Lock the interface."""
        self._lock.acquire()
    
    def unlock(self):
        """! @brief Unlock the interface."""
        self._lock.release()

    def _read_protocol_version(self):
        """! Determine the CMSIS-DAP protocol version."""
        protocol_version_str = self._protocol.dap_info(self.ID.CMSIS_DAP_PROTOCOL_VERSION)
        # Just in case we don't get a valid response, default to the lowest version (not including betas).
        if not protocol_version_str:
            self._cmsis_dap_version = CMSISDAPVersion.V1_0_0
            return

        # Convert the version to a 3-tuple for easy comparison.
        # 1.2.3 will be converted to (1,2,3), 1.10 to (1,1,0), and so on.
        #
        # There are two version formats returned from the reference CMSIS-DAP code: 2-field and 3-field.
        # The older versions return versions like "1.07" and "1.10", while recent versions return "1.2.0"
        # or "2.0.0".
        #
        # Some CMSIS-DAP compatible debug probes from various vendors return the probe's firmware version
        # rather than protocol version (like DAPLink versions 0254 and 0255 do) due to a misunderstanding
        # based on unclear documentation. These cases are handled by the additional error checking below.
        #
        # Note that the exact version identified here is not that important, as it's not used much in
        # this code (so far at least). There are also DAP_Info Capability bits for availability of certain
        # commands that should be used instead of checking the version.
        try:
            fw_version = protocol_version_str.split('.')
            major = int(fw_version[0])
            # Handle version of the form "1.10" by treating the two digits after the dot as minor and patch.
            if (len(fw_version) == 2) and len(fw_version[1]) == 2:
                minor = int(fw_version[1][0])
                patch = int(fw_version[1][1])
            # All other forms.
            else:
                minor = int(fw_version[1] if len(fw_version) > 1 else 0)
                patch = int(fw_version[2] if len(fw_version) > 2 else 0)
            self._cmsis_dap_version = (major, minor, patch)
        except ValueError:
            # One of the protocol version fields had a non-numeric character, indicating it is not a valid
            # CMSIS-DAP version number. Default to the lowest version.
            self._cmsis_dap_version = CMSISDAPVersion.V1_0_0
        
        # Validate the version against known CMSIS-DAP minor versions. This will also catch the beta release
        # versions of CMSIS-DAP, 0.01 and 0.02, and raise them to 1.0.0.
        if self._cmsis_dap_version[:2] not in CMSISDAPVersion.minor_versions():
            self._cmsis_dap_version = CMSISDAPVersion.V1_0_0

    @locked
    def open(self):
        if self._interface is None:
            raise DAPAccessIntf.DeviceError("Unable to open device with no interface")

        self._interface.open()
        self._protocol = CMSISDAPProtocol(self._interface)

        if session.Session.get_current().options['cmsis_dap.limit_packets'] or DAPSettings.limit_packets:
            self._packet_count = 1
            LOG.debug("Limiting packet count to %d", self._packet_count)
        else:
            self._packet_count = self._protocol.dap_info(self.ID.MAX_PACKET_COUNT)

        # Get the protocol version.
        self._read_protocol_version()

        # Read the firmware version if the protocol supports it.
        # THe PRODUCT_FW_VERSION ID was added in versions 1.3.0 (HID) and 2.1.0 (bulk).
        if (self._cmsis_dap_version >= CMSISDAPVersion.V2_1_0) or (self._cmsis_dap_version >= CMSISDAPVersion.V1_3_0
                and self._cmsis_dap_version < CMSISDAPVersion.V2_0_0):
            self._fw_version = self._protocol.dap_info(self.ID.PRODUCT_FW_VERSION)
        
        # Log probe's firmware version.
        if self._fw_version:
            LOG.debug("CMSIS-DAP probe %s: firmware version %s, protocol version %i.%i.%i",
                    self._unique_id, self._fw_version, *self._cmsis_dap_version)
        else:
            LOG.debug("CMSIS-DAP probe %s: protocol version %i.%i.%i", self._unique_id, *self._cmsis_dap_version)

        self._interface.set_packet_count(self._packet_count)
        self._packet_size = self._protocol.dap_info(self.ID.MAX_PACKET_SIZE)
        self._interface.set_packet_size(self._packet_size)
        self._capabilities = self._protocol.dap_info(self.ID.CAPABILITIES)
        self._has_swo_uart = (self._capabilities & Capabilities.SWO_UART) != 0
        if self._has_swo_uart:
            self._swo_buffer_size = self._protocol.dap_info(self.ID.SWO_BUFFER_SIZE)
        else:
            self._swo_buffer_size = 0
        self._swo_status = SWOStatus.DISABLED

        self._init_deferred_buffers()

    @locked
    def close(self):
        assert self._interface is not None
        self.flush()
        self._interface.close()

    def get_unique_id(self):
        return self._unique_id

    @locked
    def assert_reset(self, asserted):
        self.flush()
        if asserted:
            self._protocol.set_swj_pins(0, Pin.nRESET)
        else:
            self._protocol.set_swj_pins(Pin.nRESET, Pin.nRESET)
    
    @locked
    def is_reset_asserted(self):
        self.flush()
        pins = self._protocol.set_swj_pins(0, Pin.NONE)
        return (pins & Pin.nRESET) == 0

    @locked
    def set_clock(self, frequency):
        self.flush()
        self._protocol.set_swj_clock(frequency)
        self._frequency = frequency

    def get_swj_mode(self):
        return self._dap_port

    def set_deferred_transfer(self, enable):
        """! @brief Allow transfers to be delayed and buffered

        By default deferred transfers are turned on.  When off, all reads and
        writes will be completed by the time the function returns.

        When enabled packets are buffered and sent all at once, which
        increases speed.  When memory is written to, the transfer
        might take place immediately, or might take place on a future
        memory write.  This means that an invalid write could cause an
        exception to occur on a later, unrelated write.  To guarantee
        that previous writes are complete call the flush() function.
        """
        if self._deferred_transfer and not enable:
            self.flush()
        self._deferred_transfer = enable

    @locked
    def flush(self):
        # Send current packet
        self._send_packet()
        # Read all backlogged
        for _ in range(len(self._commands_to_read)):
            self._read_packet()

    @locked
    def identify(self, item):
        assert isinstance(item, DAPAccessIntf.ID)
        self.flush()
        return self._protocol.dap_info(item)

    @locked
    def vendor(self, index, data=None):
        if data is None:
            data = []
        self.flush()
        return self._protocol.vendor(index, data)

    # ------------------------------------------- #
    #             Target access functions
    # ------------------------------------------- #
    @locked
    def connect(self, port=DAPAccessIntf.PORT.DEFAULT):
        assert isinstance(port, DAPAccessIntf.PORT)
        actual_port = self._protocol.connect(port.value)
        self._dap_port = DAPAccessIntf.PORT(actual_port)
        # set clock frequency
        self._protocol.set_swj_clock(self._frequency)
        # configure transfer
        self._protocol.transfer_configure()
        
        # configure the selected protocol with defaults.
        if self._dap_port == DAPAccessIntf.PORT.SWD:
            self.configure_swd()
        elif self._dap_port == DAPAccessIntf.PORT.JTAG:
            self.configure_jtag()

    @locked
    def configure_swd(self, turnaround=1, always_send_data_phase=False):
        self.flush()
        self._protocol.swd_configure(turnaround, always_send_data_phase)
    
    @locked
    def configure_jtag(self, devices_irlen=None):
        self.flush()
        self._protocol.jtag_configure(devices_irlen)

    @locked
    def swj_sequence(self, length, bits):
        self.flush()
        self._protocol.swj_sequence(length, bits)

    @locked
    def swd_sequence(self, sequences):
        self.flush()
        self._protocol.swd_sequence(sequences)

    @locked
    def jtag_sequence(self, cycles, tms, read_tdo, tdi):
        self.flush()
        return self._protocol.jtag_sequence(cycles, tms, read_tdo, tdi)

    @locked
    def disconnect(self):
        self.flush()
        self._protocol.disconnect()
    
    def has_swo(self):
        return self._has_swo_uart
    
    @locked
    def swo_configure(self, enabled, rate):
        self.flush()

        # Don't send any commands if the SWO commands aren't supported.
        if not self._has_swo_uart:
            return False
        
        # Before we attempt any configuration, we must explicitly disable SWO
        # (if SWO is enabled, setting any other configuration fails).
        self._swo_disable()

        try:
            if enabled:
                # Select the streaming SWO endpoint if available.
                if self._interface.has_swo_ep:
                    transport = DAPSWOTransport.DAP_SWO_EP
                else:
                    transport = DAPSWOTransport.DAP_SWO_DATA
                
                if self._protocol.swo_transport(transport) != 0:
                    self._swo_disable()
                    return False
                if self._protocol.swo_mode(DAPSWOMode.UART) != 0:
                    self._swo_disable()
                    return False
                if self._protocol.swo_baudrate(rate) == 0:
                    self._swo_disable()
                    return False
                self._swo_status = SWOStatus.CONFIGURED

            return True
        except DAPAccessIntf.CommandError as e:
            LOG.debug("Exception while configuring SWO: %s", e)
            self._swo_disable()
            return False
    
    # Doesn't need @locked because it is only called from swo_configure().
    def _swo_disable(self):
        try:
            self._protocol.swo_mode(DAPSWOMode.OFF)
            self._protocol.swo_transport(DAPSWOTransport.NONE)
        except DAPAccessIntf.CommandError as e:
            LOG.debug("Exception while disabling SWO: %s", e)
        finally:
            self._swo_status = SWOStatus.DISABLED
    
    @locked
    def swo_control(self, start):
        self.flush()

        # Don't send any commands if the SWO commands aren't supported.
        if not self._has_swo_uart:
            return False
        
        if start:
            self._protocol.swo_control(DAPSWOControl.START)
            if self._interface.has_swo_ep:
                self._interface.start_swo()
            self._swo_status = SWOStatus.RUNNING
        else:
            self._protocol.swo_control(DAPSWOControl.STOP)
            if self._interface.has_swo_ep:
                self._interface.stop_swo()
            self._swo_status = SWOStatus.CONFIGURED
        return True
    
    @locked
    def get_swo_status(self):
        return self._protocol.swo_status()
    
    def swo_read(self, count=None):
        # The separate SWO EP can be read without locking.
        if self._interface.has_swo_ep:
            return self._interface.read_swo()
        else:
            if count is None:
                count = self._packet_size
            # Must lock and flush the command queue since we're using the SWO read command that shares
            # the command EP.
            with self._lock:
                self.flush()
                status, count, data = self._protocol.swo_data(count)
                return bytearray(data)

    def write_reg(self, reg_id, value, dap_index=0):
        assert reg_id in self.REG
        assert isinstance(value, int)
        assert isinstance(dap_index, int)

        request = WRITE
        if reg_id.value < 4:
            request |= DP_ACC
        else:
            request |= AP_ACC
        request |= (reg_id.value % 4) * 4
        self._write(dap_index, 1, request, [value])

    def read_reg(self, reg_id, dap_index=0, now=True):
        assert reg_id in self.REG
        assert isinstance(dap_index, int)
        assert isinstance(now, bool)

        request = READ
        if reg_id.value < 4:
            request |= DP_ACC
        else:
            request |= AP_ACC
        request |= (reg_id.value % 4) << 2
        transfer = self._write(dap_index, 1, request, None)
        assert transfer is not None

        def read_reg_cb():
            res = transfer.get_result()
            assert len(res) == 1
            res = res[0]
            return res

        if now:
            return read_reg_cb()
        else:
            return read_reg_cb

    def reg_write_repeat(self, num_repeats, reg_id, data_array, dap_index=0):
        assert isinstance(num_repeats, int)
        assert num_repeats == len(data_array)
        assert reg_id in self.REG
        assert isinstance(dap_index, int)

        request = WRITE
        if reg_id.value < 4:
            request |= DP_ACC
        else:
            request |= AP_ACC
        request |= (reg_id.value % 4) * 4
        self._write(dap_index, num_repeats, request, data_array)

    def reg_read_repeat(self, num_repeats, reg_id, dap_index=0,
                        now=True):
        assert isinstance(num_repeats, int)
        assert reg_id in self.REG
        assert isinstance(dap_index, int)
        assert isinstance(now, bool)

        request = READ
        if reg_id.value < 4:
            request |= DP_ACC
        else:
            request |= AP_ACC
        request |= (reg_id.value % 4) * 4
        transfer = self._write(dap_index, num_repeats, request, None)
        assert transfer is not None

        def reg_read_repeat_cb():
            res = transfer.get_result()
            assert len(res) == num_repeats
            return res

        if now:
            return reg_read_repeat_cb()
        else:
            return reg_read_repeat_cb
    # ------------------------------------------- #
    #          Private functions
    # ------------------------------------------- #

    def _init_deferred_buffers(self):
        """! @brief Initialize or reinitialize all the deferred transfer buffers

        Calling this method will drop all pending transactions
        so use with care.
        """
        # List of transfers that have been started, but
        # not completed (started by write_reg, read_reg,
        # reg_write_repeat and reg_read_repeat)
        self._transfer_list = collections.deque()
        # The current packet - this can contain multiple
        # different transfers
        self._crnt_cmd = _Command(self._packet_size)
        # Packets that have been sent but not read
        self._commands_to_read = collections.deque()
        # Buffer for data returned for completed commands.
        # This data will be added to transfers
        self._command_response_buf = bytearray()

    @locked
    def _read_packet(self):
        """! @brief Reads and decodes a single packet

        Reads a single packet from the device and
        stores the data from it in the current Command
        object
        """
        # Grab command, send it and decode response
        cmd = self._commands_to_read.popleft()
        try:
            raw_data = self._interface.read()
            raw_data = bytearray(raw_data)
            decoded_data = cmd.decode_data(raw_data)
        except Exception as exception:
            self._abort_all_transfers(exception)
            raise

        decoded_data = bytearray(decoded_data)
        self._command_response_buf.extend(decoded_data)

        # Attach data to transfers
        pos = 0
        while True:
            size_left = len(self._command_response_buf) - pos
            if size_left == 0:
                # If size left is 0 then the transfer list might
                # be empty, so don't try to access element 0
                break
            transfer = self._transfer_list[0]
            size = transfer.get_data_size()
            if size > size_left:
                break

            self._transfer_list.popleft()
            data = self._command_response_buf[pos:pos + size]
            pos += size
            transfer.add_response(data)

        # Remove used data from _command_response_buf
        if pos > 0:
            self._command_response_buf = self._command_response_buf[pos:]

    @locked
    def _send_packet(self):
        """! @brief Send a single packet to the interface

        This function guarantees that the number of packets
        that are stored in daplink's buffer (the number of
        packets written but not read) does not exceed the
        number supported by the given device.
        """
        cmd = self._crnt_cmd
        if cmd.get_empty():
            return

        max_packets = self._interface.get_packet_count()
        if len(self._commands_to_read) >= max_packets:
            self._read_packet()
        data = cmd.encode_data()
        try:
            self._interface.write(list(data))
        except Exception as exception:
            self._abort_all_transfers(exception)
            raise
        self._commands_to_read.append(cmd)
        self._crnt_cmd = _Command(self._packet_size)

    @locked
    def _write(self, dap_index, transfer_count,
               transfer_request, transfer_data):
        """! @brief Write one or more commands
        """
        assert dap_index == 0  # dap index currently unsupported
        assert isinstance(transfer_count, int)
        assert isinstance(transfer_request, int)
        assert transfer_data is None or len(transfer_data) > 0

        # Create transfer and add to transfer list
        transfer = None
        if transfer_request & READ:
            transfer = _Transfer(self, dap_index, transfer_count,
                                 transfer_request, transfer_data)
            self._transfer_list.append(transfer)

        # Build physical packet by adding it to command
        cmd = self._crnt_cmd
        size_to_transfer = transfer_count
        trans_data_pos = 0
        while size_to_transfer > 0:
            # Get the size remaining in the current packet for the given request.
            size = cmd.get_request_space(size_to_transfer, transfer_request, dap_index)

            # This request doesn't fit in the packet so send it.
            if size == 0:
                TRACE.debug("_write: send packet [size==0]")
                self._send_packet()
                cmd = self._crnt_cmd
                continue

            # Add request to packet.
            if transfer_data is None:
                data = None
            else:
                data = transfer_data[trans_data_pos:trans_data_pos + size]
            cmd.add(size, transfer_request, data, dap_index)
            size_to_transfer -= size
            trans_data_pos += size

            # Packet has been filled so send it
            if cmd.get_full():
                TRACE.debug("_write: send packet [full]")
                self._send_packet()
                cmd = self._crnt_cmd

        if not self._deferred_transfer:
            self.flush()

        return transfer

    @locked
    def _abort_all_transfers(self, exception):
        """! @brief Abort any ongoing transfers and clear all buffers
        """
        pending_reads = len(self._commands_to_read)
        # invalidate _transfer_list
        for transfer in self._transfer_list:
            transfer.add_error(exception)
        # clear all deferred buffers
        self._init_deferred_buffers()
        # finish all pending reads and ignore the data
        # Only do this if the error is a transfer error.
        # Otherwise this could cause another exception
        if isinstance(exception, DAPAccessIntf.TransferError):
            for _ in range(pending_reads):
                self._interface.read()



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ..core import exceptions
from ..commands.base import CommandBase

LOG = logging.getLogger(__name__)

class ThreadsCommand(CommandBase):
    INFO = {
            'names': ['threads'],
            'group': 'gdbserver',
            'category': 'threads',
            'nargs': 1,
            'usage': "{flush,enable,disable,status}",
            'help': "Control thread awareness.",
            }
    
    def parse(self, args):
        self.action = args[0]
        if self.action not in ('flush', 'enable', 'disable', 'status'):
            raise exceptions.CommandError("invalid action")
    
    def execute(self):
        # Get the gdbserver for the selected core.
        core_number = self.context.selected_core.core_number
        try:
            gdbserver = self.context.session.gdbservers[core_number]
        except KeyError:
            raise exceptions.CommandError("no gdbserver for core #%i" % core_number)

        if gdbserver.thread_provider is None:
            self.context.write("Threads are unavailable")
            return

        if self.action == 'flush':
            gdbserver.thread_provider.invalidate()
            self.context.write("Threads flushed")
        elif self.action == 'enable':
            gdbserver.thread_provider.read_from_target = True
            self.context.write("Threads enabled")
        elif self.action == 'disable':
            gdbserver.thread_provider.read_from_target = False
            self.context.write("Threads disabled")
        elif self.action == 'status':
            self.context.write("Threads are " +
                    ("enabled" if gdbserver.thread_provider.read_from_target else "disabled"))

class ArmSemihostingCommand(CommandBase):
    INFO = {
            'names': ['arm'],
            'group': 'gdbserver',
            'category': 'semihosting',
            'nargs': 2,
            'usage': "semihosting {enable,disable}",
            'help': "Enable or disable semihosting.",
            'extra_help': "Provided for compatibility with OpenOCD. The same functionality can be achieved "
                            "by setting the 'enable_semihosting' session option.",
            }
    
    def parse(self, args):
        if args[0] != 'semihosting':
            raise exceptions.CommandError("invalid action")
        if args[1] not in ('enable', 'disable'):
            raise exceptions.CommandError("invalid action")
        self.action = args[1]
    
    def execute(self):
        enable = (self.action == 'enable')
        self.context.session.options['enable_semihosting'] = enable

class GdbserverMonitorInitCommand(CommandBase):
    """@brief 'init' command for OpenOCD compatibility.
    
    Many default gdbserver configurations send an 'init' monitor command.
    """
    INFO = {
            'names': ['init'],
            'group': 'gdbserver',
            'category': 'openocd_compatibility',
            'nargs': 2,
            'usage': "init",
            'help': "Ignored; for OpenOCD compatibility.",
            }
    
    def execute(self):
        pass



# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import argparse
import os
import sys
import traceback
import logging
from unittest import mock

from pyocd.core.helpers import ConnectHelper
from pyocd.flash.file_programmer import FileProgrammer
from pyocd.probe.pydapaccess import DAPAccess
from pyocd.utility import conversion
from pyocd.utility.mask import same
from pyocd.core.memory_map import MemoryType
from pyocd.debug.elf.elf_reader import ElfReaderContext

from test_util import (
    Test,
    TestResult,
    get_session_options,
    get_target_test_params,
    get_test_binary_path,
    PYOCD_DIR,
    binary_to_elf_file,
    )

GDB_TEST_BIN = "src/gdb_test_program/gdb_test.bin"
GDB_TEST_ELF = "src/gdb_test_program/gdb_test.elf"

class DebugContextTestResult(TestResult):
    def __init__(self):
        super(DebugContextTestResult, self).__init__(None, None, None)
        self.name = "debug_context"

class DebugContextTest(Test):
    def __init__(self):
        super(DebugContextTest, self).__init__("Debug Context Test", debug_context_test)

    def run(self, board):
        try:
            result = self.test_function(board.unique_id)
        except Exception as e:
            result = DebugContextTestResult()
            result.passed = False
            print("Exception %s when testing board %s" % (e, board.unique_id))
            traceback.print_exc(file=sys.stdout)
        result.board = board
        result.test = self
        return result

def debug_context_test(board_id):
    with ConnectHelper.session_with_chosen_probe(unique_id=board_id, **get_session_options()) as session:
        board = session.board
        target = session.target

        test_params = get_target_test_params(session)
        session.probe.set_clock(test_params['test_clock'])

        memory_map = target.get_memory_map()
        boot_region = memory_map.get_boot_memory()
        ram_region = memory_map.get_default_region_of_type(MemoryType.RAM)
        ram_base = ram_region.start
        binary_file = get_test_binary_path(board.test_binary)
        gdb_test_binary_file = os.path.join(PYOCD_DIR, GDB_TEST_BIN)

        # Read the gdb test binary file.
        with open(gdb_test_binary_file, "rb") as f:
            gdb_test_binary_data = list(bytearray(f.read()))

        # Read the test binary file.
        with open(binary_file, "rb") as f:
            test_binary_data = bytearray(f.read())
        test_binary_data_length = len(test_binary_data)
        
        # Generate ELF file from the binary test file.
        temp_test_elf_name = binary_to_elf_file(binary_file, boot_region.start)

        test_pass_count = 0
        test_count = 0
        result = DebugContextTestResult()
        
        target.reset_and_halt()
        
        # Reproduce a gdbserver failure.
        print("\n------ Test 1: Mem cache ------")
        
        ctx = target.get_target_context()

        print("Writing gdb test binary")
        ctx.write_memory_block8(ram_base, gdb_test_binary_data)
        
        print("Reading first chunk")
        data = ctx.read_memory_block8(ram_base, 64)
        if data == gdb_test_binary_data[:64]:
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")
        test_count += 1
            
        print("Reading N chunks")
        did_pass = True
        for n in range(8):
            offset = 0x7e + (4 * n)
            data = ctx.read_memory_block8(ram_base + offset, 4)
            if data == gdb_test_binary_data[offset:offset + 4]:
                test_pass_count += 1
            else:
                did_pass = False
            test_count += 1
        if did_pass:
            print("TEST PASSED")
        else:
            print("TEST FAILED")
        
        # Force a memory cache clear.
        target.step()
        
        # ELF reader test goals:
        # 1. Verify correct data is read without accessing the target memory.
        # 2. Test null interval failure.
        #
        print("\n------ Test 2: ELF reader ------")
        
        # Set the elf on the target, which will add a context to read from the elf.
        target.elf = temp_test_elf_name
        ctx = target.get_target_context()
        
        print("Check that ElfReaderContext was created")
        if isinstance(ctx, ElfReaderContext):
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")
        test_count += 1
        
        # Program the test binary.
        print("Programming test binary to boot memory")
        FileProgrammer(session).program(binary_file, base_address=boot_region.start)

        with mock.patch.object(target.selected_core, 'read_memory_block32') as read_block32_mock:
            test_len = min(4096, test_binary_data_length)
            print("Reading %d bytes of test binary from context." % test_len)
            data = ctx.read_memory_block32(boot_region.start, test_len // 4)
            data = conversion.u32le_list_to_byte_list(data)
            if same(data, test_binary_data[:test_len]):
                print("PASSED: expected data returned")
                test_pass_count += 1
            else:
                print("FAILED: unexpected data")
            test_count += 1

            # Verify the target memory wasn't accessed.
            try:
                read_block32_mock.assert_not_called()
            except AssertionError:
                print("FAILED: target memory was accessed")
            else:
                print("PASSED: target memory was not accessed")
                test_pass_count += 1
            test_count += 1

        print("\nTest Summary:")
        print("Pass count %i of %i tests" % (test_pass_count, test_count))
        if test_pass_count == test_count:
            print("DEBUG CONTEXT TEST PASSED")
        else:
            print("DEBUG CONTEXT TEST FAILED")

        # Clean up.
        target.reset()

        result.passed = test_count == test_pass_count
        return result

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='pyOCD debug context test')
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    parser.add_argument("-da", "--daparg", dest="daparg", nargs='+', help="Send setting to DAPAccess layer.")
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    DAPAccess.set_args(args.daparg)
    # Set to debug to print some of the decisions made while flashing
    session = ConnectHelper.session_with_chosen_probe(**get_session_options())
    test = DebugContextTest()
    result = [test.run(session.board)]




# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from enum import Enum

from .component import CoreSightComponent
from ..core import exceptions
from ..utility.timeout import Timeout

LOG = logging.getLogger(__name__)

class ComPortError(exceptions.Error):
    """! @brief Base class for SDC-600 exceptions."""
    pass

class UnexpectedFlagError(ComPortError):
    """! @brief Received an unexpected or out of order flag byte."""
    pass

class LinkError(ComPortError):
    """! @brief Received a link error flag (LERR)."""
    pass

class LinkClosedException(ComPortError):
    """! @brief Received an unexpected or out of order flag byte."""
    def __init__(self, phase):
        self._phase = phase
    
    @property
    def phase(self):
        """! @brief The link phase that was closed from the other side."""
        return self._phase

class SDC600(CoreSightComponent):
    """! @brief SDC-600 component.
    """
    
    ## Default timeout for an operation or packet transfer.
    TRANSFER_TIMEOUT = 30.0
    
    class LinkPhase(Enum):
        """! @brief COM Port link phases."""
        ## Hardware-defined link phase.
        PHASE1 = 1
        ## Software-defiend link phase.
        PHASE2 = 2
    
    class Register:
        """! @brief Namespace for SDC-600 register offset constants."""
        # Register offsets.
        VIDR        = 0xD00
        FIDTXR      = 0xD08
        FIDRXR      = 0xD0C
        ICSR        = 0xD10
        DR          = 0xD20
        SR          = 0xD2C
        DBR         = 0xD30
        SR_ALIAS    = 0xD3C

        # FIDTXR and FIDRXR bit definitions.
        FIDxXR_xXI_MASK     = (0x00000001)
        FIDxXR_xXI_SHIFT    = (0)
        FIDxXR_xXINT_MASK   = (0x00000002)
        FIDxXR_xXINT_SHIFT  = (1)
        FIDxXR_xXW_MASK     = (0x000000f0)
        FIDxXR_xXW_SHIFT    = (4)
        FIDxXR_xXSZ8_MASK   = (0x00000100)
        FIDxXR_xXSZ8_SHIFT  = (8)
        FIDxXR_xXSZ16_MASK  = (0x00000200)
        FIDxXR_xXSZ16_SHIFT = (9)
        FIDxXR_xXSZ32_MASK  = (0x00000400)
        FIDxXR_xXSZ32_SHIFT = (10)
        FIDxXR_xXFD_MASK    = (0x000f0000)
        FIDxXR_xXFD_SHIFT   = (16)
        
        # SR bit definitions.
        SR_TXS_MASK         = (0x000000ff)
        SR_TXS_SHIFT        = (0)
        SR_RRDIS_MASK       = (0x00001000)
        SR_RRDIS_SHIFT      = (12)
        SR_TXOE_MASK        = (0x00002000)
        SR_TXOE_SHIFT       = (13)
        SR_TXLE_MASK        = (0x00004000)
        SR_TXLE_SHIFT       = (14)
        SR_TRINPROG_MASK    = (0x00008000)
        SR_TRINPROG_SHIFT   = (18)
        SR_RXF_MASK         = (0x00ff0000)
        SR_RXF_SHIFT        = (16)
        SR_RXLE_MASK        = (0x40000000)
        SR_RXLE_SHIFT       = (30)
        SR_PEN_MASK         = (0x80000000)
        SR_PEN_SHIFT        = (31)
    
    class Flag:
        """! @brief Namespace with SDC-600 flag byte constants."""
        IDR     = 0xA0
        IDA     = 0xA1
        LPH1RA  = 0xA6
        LPH1RL  = 0xA7
        LPH2RA  = 0xA8
        LPH2RL  = 0xA9
        LPH2RR  = 0xAA
        LERR    = 0xAB
        START   = 0xAC
        END     = 0xAD
        ESC     = 0xAE
        NULL    = 0xAF
        
        # All bytes with 0b101 in bits [7:5] are flag bytes.
        MASK = 0xE0
        IDENTIFIER = 0b10100000
        
        ## Map from flag value to name.
        NAME = {
            IDR     : "IDR",
            IDA     : "IDA",
            LPH1RA  : "LPH1RA",
            LPH1RL  : "LPH1RL",
            LPH2RA  : "LPH2RA",
            LPH2RL  : "LPH2RL",
            LPH2RR  : "LPH2RR",
            LERR    : "LERR",
            START   : "START",
            END     : "END",
            ESC     : "ESC",
            NULL    : "NULL",
            }
    
    ## NULL bytes must be written to the upper bytes, and will be present in the upper bytes
    # when read.
    NULL_FILL = 0xAFAFAF00
    
    def __init__(self, ap, cmpid=None, addr=None):
        super(SDC600, self).__init__(ap, cmpid, addr)
        self._tx_width = 0
        self._rx_width = 0
        self._current_link_phase = None

    def init(self):
        """! @brief Inits the component.
        
        Reads the RX and TX widths and whether the SDC-600 is enabled. All error flags are cleared.
        """
        fidtx = self.ap.read32(self.Register.FIDTXR)
        LOG.debug("fidtx=0x%08x", fidtx)
        fidrx = self.ap.read32(self.Register.FIDRXR)
        LOG.debug("fidrx=0x%08x", fidrx)
        
        self._tx_width = (fidtx & self.Register.FIDxXR_xXW_MASK) >> self.Register.FIDxXR_xXW_SHIFT
        
        self._rx_width = (fidrx & self.Register.FIDxXR_xXW_MASK) >> self.Register.FIDxXR_xXW_SHIFT
        
        status = self.ap.read32(self.Register.SR)
        LOG.debug("status=0x%08x", status)
        self._is_enabled = (status & self.Register.SR_PEN_MASK) != 0
        
        # Clear any error flags.
        error_flags = status & (self.Register.SR_TXOE_MASK | self.Register.SR_TXLE_MASK)
        if error_flags:
            self.ap.write32(self.Register.SR, error_flags)
    
    @property
    def is_enabled(self):
        """! @brief Whether the SDC-600 is enabled."""
        return self._is_enabled
    
    @property
    def is_reboot_request_enabled(self):
        """! @brief Whether the Reboot Request feature is enabled in the SDC-600."""
        return (self.ap.read32(self.Register.SR) & self.Register.SR_RRDIS_MASK) == 0
    
    @property
    def current_link_phase(self):
        """! @brief Currently established link phase.
        @return Either None or one of the SDC600.LinkPhase enums.
        """
        return self._current_link_phase

    def _read1(self, to_):
        """! @brief Read a single byte.
        
        If a NULL byte is received, it is ignored and another byte is read. No other flag bytes
        are processed.
        
        @exception TimeoutError
        """
        while True:
            # Wait until a byte is ready in the receive FIFO.
            while to_.check():
                if (self.ap.read32(self.Register.SR) & self.Register.SR_RXF_MASK) != 0:
                    break
            else:
                raise exceptions.TimeoutError("timeout while reading from SDC-600")

            # Read the data register and strip off NULL bytes in high bytes.
            value = self.ap.read32(self.Register.DR) & 0xFF

            # Ignore NULL flag bytes.
            if value == self.Flag.NULL:
                continue
            
            return value
        
    def _write1(self, value, to_):
        """! @brief Write one or more bytes.
        @exception TimeoutError
        """
        # Wait until room is available in the transmit FIFO.
        while to_.check():
            if (self.ap.read32(self.Register.SR) & self.Register.SR_TXS_MASK) != 0:
                break
        else:
            raise exceptions.TimeoutError("timeout while writing to SDC-600")

        # Write this byte to the transmit FIFO.
        dbr_value = self.NULL_FILL | (value & 0xFF)
        self.ap.write32(self.Register.DR, dbr_value)

    def _check_flags(self, value, to_):
        """! @brief Handle link and error related flag bytes.
        @param self
        @param value Integer byte value to check.
        @param to_ Timeout object.
        @exception UnexpectedFlagError
        @exception LinkClosedException
        @exception LinkError
        @exception TimeoutError
        """
        if value == self.Flag.LPH1RL:
            LOG.debug("got LPH1RL!")
            self._current_link_phase = None
            raise LinkClosedException(self.LinkPhase.PHASE1)
        elif value == self.Flag.LPH2RL:
            LOG.debug("got LPH2RL!")
            # Target killed the phase 2 connection. Send required reply.
            self._current_link_phase = self.LinkPhase.PHASE1
            self._write1(self.Flag.LPH2RL, to_)
            raise LinkClosedException(self.LinkPhase.PHASE2)
        elif value == self.Flag.LERR:
            LOG.debug("got LERR!")
            raise LinkError()
        # Catch reserved flags.
        elif (0xA2 <= value <= 0xA5) or (0xB0 <= value <= 0xBF):
            raise UnexpectedFlagError("received reserved flag value ({:#04x})".format(value))

    def _expect_flag(self, flag, to_):
        """! @brief Read a byte and compare to expected value.
        @param self
        @param flag Integer flag byte value to match.
        @param to_ Timeout object.
        @exception UnexpectedFlagError
        @exception LinkClosedException
        @exception TimeoutError
        """
        value = self._read1(to_)
        if value != flag:
            # Check certain flags we have to handle. This will raise if a flag is handled.
            self._check_flags(value, to_)
            # _check_flags() did not raise, so we should .
            raise UnexpectedFlagError("got {:#04x} instead of expected {} ({:#04x})".format(
                        value, self.Flag.NAME[flag], flag))
        else:
            LOG.debug("got expected %s", self.Flag.NAME[value])

    def _stuff(self, data):
        """! @brief Perform COM Encapsulation byte stuffing.
        @param self
        @param data List of integers of the original data.
        @return List of integers for the escaped version of _data_.
        """
        result = []
        for value in data:
            # Values matching flag bytes just get copied to output.
            if (value & self.Flag.MASK) == self.Flag.IDENTIFIER:
                # Insert escape flag.
                result.append(self.Flag.ESC)
                
                # Invert high bit.
                value ^= 0x80
            
            result.append(value)
        return result

    def _destuff(self, data):
        """! @brief Remove COM Encapsulation byte stuffing.
        @param self
        @param data List of integers. The only acceptable flag byte is ESC.
        @return List of integers properly de-stuffed.
        """
        result = []
        i = 0
        while i < len(data):
            value = data[i]
            
            # Check for escaped bytes.
            if value == self.Flag.ESC:
                # Skip over escape.
                i += 1
                
                # Get escaped byte and invert high bit to destuff it.
                value = data[i] ^ 0x80
            
            result.append(value)
            
            i += 1
        return result

    def _read_packet_data_to_end(self, to_):
        """! @brief Read an escaped packet from the first message byte to the end.
        @exception UnexpectedFlagError
        @exception LinkClosedException
        @exception TimeoutError
        """
        result = []
        while to_.check():
            value = self._read1(to_)
            
            # Check for the packet end marker flag.
            if value == self.Flag.END:
                break
            # Handle other flag bytes. This will raise on any detected flags.
            elif (value & self.Flag.MASK) == self.Flag.IDENTIFIER:
                self._check_flags(value, to_)

            # Append data bytes.
            result.append(value)
        else:
            raise exceptions.TimeoutError("timeout while reading from SDC-600")
        
        return self._destuff(result)

    def receive_packet(self, timeout=TRANSFER_TIMEOUT):
        """! @brief Read a data packet.
        
        Reads a packet (PDU) from the target and removes byte stuffing. The timeout for reading the
        entire packet can be set via the _timeout_ parameter.
        
        As data is read from the target, special flags for link errors or to close either phase of
        the link are handled and an appropriate exception is raised.
        
        The connection must be in link phase 2.
        
        @param self
        @param timeout Optional timeout for reading the entire packet. If reading times out, a
            TimeoutError exception is raised.
        @return List of integer byte values of the de-escaped packet contents.
        
        @exception UnexpectedFlagError
        @exception LinkClosedException
        @exception TimeoutError
        """
        assert self._current_link_phase == self.LinkPhase.PHASE2
        with Timeout(timeout) as to_:
            self._expect_flag(self.Flag.START, to_)
            return self._read_packet_data_to_end(to_)
    
    def send_packet(self, data, timeout=TRANSFER_TIMEOUT):
        """! @brief Send a data packet.
        
        Sends the provided data to the target as a single packet (PDU), escaping bytes as necessary.
        No data is read while the packet is sent, so if the target closes the connection it will
        not be detected.
        
        The connection must be in link phase 2.
        
        @param self
        @param data List of integer byte values to send. Must not be pre-escaped.
        @param timeout Optional timeout for reading the entire packet. If reading times out, a
            TimeoutError exception is raised.
        
        @exception UnexpectedFlagError
        @exception TimeoutError
        """
        assert self._current_link_phase == self.LinkPhase.PHASE2
        with Timeout(timeout) as to_:
            self._write1(self.Flag.START, to_)
            for value in self._stuff(data):
                self._write1(value, to_)
            self._write1(self.Flag.END, to_)
    
    def open_link(self, phase, timeout=TRANSFER_TIMEOUT):
        """! @brief Send the LPH1RA or LPH2RA flag.
        @exception UnexpectedFlagError
        @exception LinkClosedException
        @exception TimeoutError
        """
        with Timeout(timeout) as to_:
            if phase == self.LinkPhase.PHASE1:
                assert self._current_link_phase is None

                # Close link phase 1 first, to put it in a known state.
                self.close_link(self.LinkPhase.PHASE1)
        
                LOG.debug("sending LPH1RA")
                self._write1(self.Flag.LPH1RA, to_)
                self._expect_flag(self.Flag.LPH1RA, to_)
                
                self._current_link_phase = self.LinkPhase.PHASE1
            elif phase == self.LinkPhase.PHASE2:
                assert self._current_link_phase == self.LinkPhase.PHASE1

                LOG.debug("sending LPH2RA")
                self._write1(self.Flag.LPH2RA, to_)
                self._expect_flag(self.Flag.LPH2RA, to_)
                
                self._current_link_phase = self.LinkPhase.PHASE2
            else:
                raise ValueError("unrecognized phase value")

    def close_link(self, phase, timeout=TRANSFER_TIMEOUT):
        """! @brief Send the LPH1RL or LPH2RL flag.
        
        Link phase 1 can be closed from any state. Link phase 2 can only be closed when the
        connection is already in that phase.
        
        @exception UnexpectedFlagError
        @exception TimeoutError
        """
        with Timeout(timeout) as to_:
            if phase == self.LinkPhase.PHASE1:
                # Link phase 1 can be closed from any state, so we don't assert here.
                LOG.debug("sending LPH1RL")
                self._write1(self.Flag.LPH1RL, to_)
                self._expect_flag(self.Flag.LPH1RL, to_)
            
                self._current_link_phase = None
            elif phase == self.LinkPhase.PHASE2:
                assert self._current_link_phase == self.LinkPhase.PHASE2

                LOG.debug("sending LPH2RL")
                self._write1(self.Flag.LPH2RL, to_)
                self._expect_flag(self.Flag.LPH2RL, to_)
            
                self._current_link_phase = self.LinkPhase.PHASE1
            else:
                raise ValueError("unrecognized phase value")

    def _log_status(self):
        status = self.ap.read32(self.Register.SR)
        LOG.info("status=0x%08x phase=%s", status, self._current_link_phase)

    def read_protocol_id(self, timeout=TRANSFER_TIMEOUT):
        """! @brief Read and return the 6-byte protocol ID.
        @exception UnexpectedFlagError
        @exception LinkClosedException
        @exception TimeoutError
        """
        with Timeout(timeout) as to_:
            self._write1(self.Flag.IDR, to_)
            self._expect_flag(self.Flag.IDA, to_)
            return self._read_packet_data_to_end(to_)
    
    def send_reboot_request(self, timeout=TRANSFER_TIMEOUT):
        """! @brief Send remote reboot request."""
        with Timeout(timeout) as to_:
            self._write1(self.Flag.LPH2RR, to_)
    
    def __repr__(self):
        return "<SDC-600@{:x}: en={} txw={} rxw={} phase={}>".format(id(self),
            self._is_enabled, self._tx_width, self._rx_width, self._current_link_phase)
        


